# coding=utf-8
from __future__ import division

import os

from .analysisperiod import AnalysisPeriod
from .datacollection import HourlyContinuousCollection
from .datacollection import MonthlyCollection
from .datatype import angle, distance, energyflux, energyintensity, generic, \
    illuminance, luminance, fraction, pressure, speed, temperature
from .designday import DesignDay
from .futil import write_to_file
from .header import Header
from .location import Location
from .skymodel import calc_sky_temperature

readmode = 'rb'
try:
    from itertools import izip as zip  # python 2
except ImportError:
    xrange = range  # python 3
    readmode = 'r'


class EPW(object):
    """An EPW object containing all of the data of an .epw file.

    Args:
        file_path: Local file address to an .epw file.

    Properties:
        * location
        * annual_heating_design_day_996
        * annual_heating_design_day_990
        * annual_cooling_design_day_004
        * annual_cooling_design_day_010
        * heating_design_condition_dictionary
        * cooling_design_condition_dictionary
        * extreme_design_condition_dictionary
        * extreme_hot_weeks
        * extreme_cold_weeks
        * typical_weeks
        * monthly_ground_temperature
        * header

        * years
        * dry_bulb_temperature
        * dew_point_temperature
        * relative_humidity
        * atmospheric_station_pressure
        * extraterrestrial_horizontal_radiation
        * extraterrestrial_direct_normal_radiation
        * horizontal_infrared_radiation_intensity
        * global_horizontal_radiation
        * direct_normal_radiation
        * diffuse_horizontal_radiation
        * global_horizontal_illuminance
        * direct_normal_illuminance
        * diffuse_horizontal_illuminance
        * zenith_luminance
        * wind_direction
        * wind_speed
        * total_sky_cover
        * opaque_sky_cover
        * visibility
        * ceiling_height
        * present_weather_observation
        * present_weather_codes
        * precipitable_water
        * aerosol_optical_depth
        * snow_depth
        * days_since_last_snowfall
        * albedo
        * liquid_precipitation_depth
        * liquid_precipitation_quantity
        * sky_temperature
    """
    __slots__ = ('_file_path', '_is_header_loaded', '_is_data_loaded', '_is_ip',
                 '_data','_metadata', '_heating_dict', '_cooling_dict', '_extremes_dict',
                 '_extreme_hot_weeks', '_extreme_cold_weeks', '_typical_weeks',
                 '_monthly_ground_temps', '_is_leap_year', 'daylight_savings_start',
                 'daylight_savings_end', '_num_of_fields', 'comments_1', 'comments_2',
                 '_location', '_header')

    def __init__(self, file_path):
        """Initialize an EPW object from from a local .epw file.
        """
        self._file_path = os.path.normpath(file_path) if file_path is not None else None
        self._is_header_loaded = False
        self._is_data_loaded = False
        self._is_ip = False  # track if collections have been converted to IP

        # placeholders for the EPW data that will be imported
        self._data = []
        self._metadata = {}
        self._heating_dict = {}
        self._cooling_dict = {}
        self._extremes_dict = {}
        self._extreme_hot_weeks = {}
        self._extreme_cold_weeks = {}
        self._typical_weeks = {}
        self._monthly_ground_temps = {}
        self._is_leap_year = False
        self.daylight_savings_start = '0'
        self.daylight_savings_end = '0'
        self.comments_1 = ''
        self.comments_2 = ''
        self._num_of_fields = 35  # it is 35 for TMY3 files

    @classmethod
    def from_missing_values(cls, is_leap_year=False):
        """Initialize an EPW object with all data missing or empty.

        Note that this classmethod is intended for workflows where one plans
        to set all of the data within the EPW object.  The EPW file written
        out from the use of this method is not simulate-abe or useful since
        all hourly data slots just possess the missing value for that data
        type. To obtain a EPW that is simulate-able in EnergyPlus, one must
        at least set the following properties:

        * location
        * dry_bulb_temperature
        * dew_point_temperature
        * relative_humidity
        * atmospheric_station_pressure
        * direct_normal_radiation
        * diffuse_horizontal_radiation
        * wind_direction
        * wind_speed
        * total_sky_cover
        * opaque_sky_cover or horizontal_infrared_radiation_intensity

        Args:
            is_leap_year: A boolean to set whether the EPW object is for a leap year.

        Usage:

        .. code-block:: python

            from ladybug.epw import EPW
            from ladybug.location import Location
            epw = EPW.from_missing_values()
            epw.location = Location('Denver Golden','CO','USA',39.74,-105.18,-7.0,1829.0)
            epw.dry_bulb_temperature.values = [20] * 8760
        """
        # Initialize the class with all data missing
        epw_obj = cls(None)
        epw_obj._is_leap_year = is_leap_year
        epw_obj._location = Location()

        # create an annual analysis period
        analysis_period = AnalysisPeriod(is_leap_year=is_leap_year)

        # create headers and an empty list for each field in epw file
        headers = []
        for field_number in xrange(epw_obj._num_of_fields):
            field = EPWFields.field_by_number(field_number)
            header = Header(data_type=field.name, unit=field.unit,
                            analysis_period=analysis_period)
            headers.append(header)
            epw_obj._data.append([])

        # fill in missing datetime values and uncertainty flags.
        uncertainty = '?9?9?9?9E0?9?9?9?9?9?9?9?9?9?9?9?9?9?9?9*9*9?9?9?9'
        for dt in analysis_period.datetimes:
            hr = dt.hour if dt.hour != 0 else 24
            epw_obj._data[0].append(dt.year)
            epw_obj._data[1].append(dt.month)
            epw_obj._data[2].append(dt.day)
            epw_obj._data[3].append(hr)
            epw_obj._data[4].append(0)
            epw_obj._data[5].append(uncertainty)

        # generate missing hourly data
        calc_length = len(analysis_period.datetimes)
        for field_number in xrange(6, epw_obj._num_of_fields):
            field = EPWFields.field_by_number(field_number)
            mis_val = field.missing if field.missing is not None else 0
            for dt in xrange(calc_length):
                epw_obj._data[field_number].append(mis_val)

        # finally, build the data collection objects from the headers and data
        for i in xrange(epw_obj._num_of_fields):
            epw_obj._data[i] = HourlyContinuousCollection(headers[i], epw_obj._data[i])

        epw_obj._is_header_loaded = True
        epw_obj._is_data_loaded = True
        return epw_obj

    @classmethod
    def from_dict(cls, data):
        """ Create EPW from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

                {
                "location": {} ,  # ladybug location schema
                "data_collections": [],  # list of hourly annual hourly data collection
                    # schemas for each of the 35 fields within the EPW file.
                "metadata": {},  # dict of metadata assigned to all data collections
                "heating_dict": {},  # dict containing heating design conditions
                "cooling_dict": {},  # dict containing cooling design conditions
                "extremes_dict": {},  # dict containing extreme design conditions
                "extreme_hot_weeks": {},  # dict with values of week-long ladybug
                    # analysis period schemas signifying extreme hot weeks.
                "extreme_cold_weeks": {},  # dict with values of week-long ladybug
                    # analysis period schemas signifying extreme cold weeks.
                "typical_weeks": {},  # dict with values of week-long ladybug
                    # analysis period schemas signifying typical weeks.
                "monthly_ground_temps": {},  # dict with keys as floats signifying
                    # depths in meters below ground and values of monthly
                    # collection schema
                "is_ip": False  # Boolean // denote whether the data is in IP units
                "is_leap_year": False  # Boolean, denote whether data is for
                                       # a leap year
                "daylight_savings_start": 0,  # signify when daylight savings starts
                                              # or 0 for no daylight savings
                "daylight_savings_end" 0,  # signify when daylight savings ends
                                           # or 0 for no daylight savings
                "comments_1": ""  # String, epw comments
                "comments_2": ""  # String, epw comments
                }
        """
        # Initialize the class with all data missing
        epw_obj = cls(None)
        epw_obj._is_header_loaded = True
        epw_obj._is_data_loaded = True

        # Check required and optional keys
        required_keys = ('location', 'data_collections')
        option_keys_dict = ('metadata', 'heating_dict', 'cooling_dict',
                            'extremes_dict', 'extreme_hot_weeks', 'extreme_cold_weeks',
                            'typical_weeks', 'monthly_ground_temps')
        for key in required_keys:
            assert key in data, 'Required key "{}" is missing!'.format(key)
        assert len(data['data_collections']) == epw_obj._num_of_fields, \
            'The number of data_collections must be {}. Got {}.'.format(
                epw_obj._num_of_fields, len(data['data_collections']))
        for key in option_keys_dict:
            if key not in data:
                data[key] = {}

        # Set the required properties of the EPW object.
        epw_obj._location = Location.from_dict(data['location'])
        epw_obj._data = [HourlyContinuousCollection.from_dict(dc)
                         for dc in data['data_collections']]
        if 'is_leap_year' in data:
            epw_obj._is_leap_year = data['is_leap_year']
        if 'is_ip' in data:
            epw_obj._is_ip = data['is_ip']

        # Check that the required properties all make sense.
        for dc in epw_obj._data:
            assert isinstance(dc, HourlyContinuousCollection), 'data_collections must ' \
                'be of HourlyContinuousCollection schema. Got {}'.format(type(dc))
            assert dc.header.analysis_period.is_annual, 'data_collections ' \
                'analysis_period must be annual.'
            assert dc.header.analysis_period.is_leap_year == epw_obj._is_leap_year, \
                'data_collections is_leap_year is not aligned with that of the EPW.'

        # Set all of the header properties if they exist in the dictionary.
        epw_obj._metadata = data['metadata']
        epw_obj.heating_design_condition_dictionary = data['heating_dict']
        epw_obj.cooling_design_condition_dictionary = data['cooling_dict']
        epw_obj.extreme_design_condition_dictionary = data['extremes_dict']

        def _dedict(parent_dict, obj):
            new_dict = {}
            for key, val in parent_dict.items():
                new_dict[key] = obj.from_dict(val)
            return new_dict
        epw_obj.extreme_hot_weeks = _dedict(data['extreme_hot_weeks'], AnalysisPeriod)
        epw_obj.extreme_cold_weeks = _dedict(data['extreme_cold_weeks'], AnalysisPeriod)
        epw_obj.typical_weeks = _dedict(data['typical_weeks'], AnalysisPeriod)
        epw_obj.monthly_ground_temperature = _dedict(
            data['monthly_ground_temps'], MonthlyCollection)

        if 'daylight_savings_start' in data:
            epw_obj.daylight_savings_start = data['daylight_savings_start']
        if 'daylight_savings_end' in data:
            epw_obj.daylight_savings_end = data['daylight_savings_end']
        if 'comments_1' in data:
            epw_obj.comments_1 = data['comments_1']
        if 'comments_2' in data:
            epw_obj.comments_2 = data['comments_2']

        return epw_obj

    @property
    def file_path(self):
        """Get path to epw file."""
        return self._file_path

    @property
    def is_header_loaded(self):
        """Return True if location data is loaded."""
        return self._is_header_loaded

    @property
    def is_data_loaded(self):
        """Return True if weather data is loaded."""
        return self._is_data_loaded

    @property
    def is_ip(self):
        """Returns True if the data collections of this file are in IP units."""
        return self._is_ip

    @property
    def location(self):
        """Return location data."""
        self._load_header_check()
        return self._location

    @location.setter
    def location(self, loc):
        self._load_header_check()
        assert isinstance(loc, Location), 'location' \
            'must be a Location object. Got {}.'.format(type(loc))
        self._location = loc

    @property
    def metadata(self):
        """Dictionary of metadata written to DataCollection headers.

        Keys typically include "source", "country", and "city").
        """
        self._load_header_check()
        return self._metadata

    @metadata.setter
    def metadata(self, meta_d):
        self._load_header_check()
        assert isinstance(meta_d, dict), 'metadata' \
            'must be a dictionary. Got {}.'.format(type(meta_d))
        self._metadata = meta_d
        for coll in self._data:
            coll.header._metadata = meta_d

    @property
    def annual_heating_design_day_996(self):
        """A design day object representing the annual 99.6% heating design day."""
        self._load_header_check()
        if bool(self._heating_dict):
            avg_press = self.atmospheric_station_pressure.average
            avg_press = None if avg_press == 999999 else avg_press
            return DesignDay.from_ashrae_dict_heating(
                self._heating_dict, self.location, False, avg_press)
        else:
            return None

    @property
    def annual_heating_design_day_990(self):
        """A design day object representing the annual 99.0% heating design day."""
        self._load_header_check()
        if bool(self._heating_dict):
            avg_press = self.atmospheric_station_pressure.average
            avg_press = None if avg_press == 999999 else avg_press
            return DesignDay.from_ashrae_dict_heating(
                self._heating_dict, self.location, True, avg_press)
        else:
            return None

    @property
    def annual_cooling_design_day_004(self):
        """A design day object representing the annual 0.4% cooling design day."""
        self._load_header_check()
        if bool(self._cooling_dict):
            avg_press = self.atmospheric_station_pressure.average
            avg_press = None if avg_press == 999999 else avg_press
            return DesignDay.from_ashrae_dict_cooling(
                self._cooling_dict, self.location, False, avg_press)
        else:
            return None

    @property
    def annual_cooling_design_day_010(self):
        """A design day object representing the annual 1.0% cooling design day."""
        self._load_header_check()
        if bool(self._cooling_dict):
            avg_press = self.atmospheric_station_pressure.average
            avg_press = None if avg_press == 999999 else avg_press
            return DesignDay.from_ashrae_dict_cooling(
                self._cooling_dict, self.location, True, avg_press)
        else:
            return None

    @property
    def heating_design_condition_dictionary(self):
        """Dictionary with ASHRAE HOF Climate Design Data for heating conditions."""
        self._load_header_check()
        return self._heating_dict

    @heating_design_condition_dictionary.setter
    def heating_design_condition_dictionary(self, des_dict):
        self._load_header_check()
        self._des_dict_check(des_dict, DesignDay.HEATING_KEYS,
                             'heating_design_condition_dictionary')
        self._heating_dict = des_dict

    @property
    def cooling_design_condition_dictionary(self):
        """Dictionary with ASHRAE HOF Climate Design Data for cooling conditions."""
        self._load_header_check()
        return self._cooling_dict

    @cooling_design_condition_dictionary.setter
    def cooling_design_condition_dictionary(self, des_dict):
        self._load_header_check()
        self._des_dict_check(des_dict, DesignDay.COOLING_KEYS,
                             'cooling_design_condition_dictionary')
        self._cooling_dict = des_dict

    @property
    def extreme_design_condition_dictionary(self):
        """Dictionary with ASHRAE HOF Climate Design Data for extreme conditions."""
        self._load_header_check()
        return self._extremes_dict

    @extreme_design_condition_dictionary.setter
    def extreme_design_condition_dictionary(self, des_dict):
        self._load_header_check()
        self._des_dict_check(des_dict, DesignDay.EXTREME_KEYS,
                             'extreme_design_condition_dictionary')
        self._extremes_dict = des_dict

    @property
    def extreme_cold_weeks(self):
        """A dictionary with AnalysisPeriods for the coldest weeks within the EPW."""
        self._load_header_check()
        self._extreme_cold_weeks.values()
        return self._extreme_cold_weeks

    @extreme_cold_weeks.setter
    def extreme_cold_weeks(self, data):
        self._load_header_check()
        self._weeks_check(data, 'extreme_cold_weeks')
        self._extreme_cold_weeks = dict(data)

    @property
    def extreme_hot_weeks(self):
        """A dictionary with AnalysisPeriods for the hottest week within the EPW."""
        self._load_header_check()
        return self._extreme_hot_weeks

    @extreme_hot_weeks.setter
    def extreme_hot_weeks(self, data):
        self._load_header_check()
        self._weeks_check(data, 'extreme_hot_weeks')
        self._extreme_hot_weeks = data

    @property
    def typical_weeks(self):
        """A dictionary with AnalysisPeriods for the typical weeks within the EPW."""
        self._load_header_check()
        return self._typical_weeks

    @typical_weeks.setter
    def typical_weeks(self, data):
        self._load_header_check()
        self._weeks_check(data, 'typical_weeks')
        self._typical_weeks = data

    @property
    def monthly_ground_temperature(self):
        """Return a dictionary of Monthly Data collections.

        The keys of this dictionary are the depths at which each set
        of temperatures occurs."""
        self._load_header_check()
        return self._monthly_ground_temps

    @monthly_ground_temperature.setter
    def monthly_ground_temperature(self, data):
        self._load_header_check()
        assert isinstance(data, dict), 'monthly_ground_temperature' \
            ' must be an OrderedDict. Got {}.'.format(type(data))
        if bool(data):
            for val in data.values():
                assert isinstance(val, MonthlyCollection), 'monthly_ground_temperature' \
                    ' must contain MonthlyCollection objects. Got {}.'.format(type(val))
        self._monthly_ground_temps = data

    @property
    def is_leap_year(self):
        """Boolean to denote whether the EPW is a leap year or not."""
        self._load_header_check()
        return self._is_leap_year

    def _load_header_check(self):
        """Check if data is loaded and, if not load it."""
        if not self.is_header_loaded:
            self._import_data(import_header_only=True)

    def _des_dict_check(self, des_dict, req_keys, cond_name):
        """Check if an input design condition dictionary is acceptable."""
        assert isinstance(des_dict, dict), '{}' \
            ' must be a dictionary. Got {}.'.format(cond_name, type(des_dict))
        if bool(des_dict):
            input_keys = list(des_dict.keys())
            for key in req_keys:
                assert key in input_keys, 'Required key "{}" was not found in ' \
                    '{}'.format(key, cond_name)

    def _weeks_check(self, data, week_type):
        """Check if input for the typical/extreme weeks of the header is correct."""
        assert isinstance(data, dict), '{}' \
            ' must be an OrderedDict. Got {}.'.format(week_type, type(data))
        if bool(data):
            for val in data.values():
                assert isinstance(val, AnalysisPeriod), '{} dictionary must contain' \
                    ' AnalysisPeriod objects. Got {}.'.format(week_type, type(val))
                assert len(val.doys_int) == 7, '{} AnalysisPeriod must be for'\
                    ' a week.  Got AnalysisPeriod for {} days.'.format(
                        week_type, type(val))

    def _import_data(self, import_header_only=False):
        """Import data from an epw file.

        Hourly data will be saved in self.data and the various header data
        will be saved in the properties above.
        """
        # perform checks on the file before opening it.
        assert os.path.isfile(self._file_path), 'Cannot find an epw file at {}'.format(
            self._file_path)
        assert self._file_path.lower().endswith('epw'), '{} is not an .epw file. \n' \
            'It does not possess the .epw file extension.'.format(self._file_path)

        with open(self._file_path, readmode) as epwin:
            line = epwin.readline()
            original_header_load = bool(self._is_header_loaded)

            if not self._is_header_loaded:
                # import location data
                # first line has location data - Here is an example
                # LOCATION,Denver Golden Nr,CO,USA,TMY3,724666,39.74,-105.18,-7.0,1829.0
                location_data = line.strip().split(',')
                self._location = Location()
                self._location.city = location_data[1].replace('\\', ' ') \
                    .replace('/', ' ')
                self._location.state = location_data[2]
                self._location.country = location_data[3]
                self._location.source = location_data[4]
                self._location.station_id = location_data[5]
                self._location.latitude = location_data[6]
                self._location.longitude = location_data[7]
                self._location.time_zone = location_data[8]
                self._location.elevation = location_data[9]

                # assemble a dictionary of metadata
                self._metadata = {
                    'source': self._location.source,
                    'country': self._location.country,
                    'city': self._location.city
                }

                self._header = [line] + [epwin.readline() for i in xrange(7)]

                # parse the heating, cooling and extreme design conditions.
                dday_data = self._header[1].strip().split(',')
                if len(dday_data) >= 2 and int(dday_data[1]) == 1:
                    if dday_data[4] == 'Heating':
                        for key, val in zip(DesignDay.HEATING_KEYS, dday_data[5:20]):
                            self._heating_dict[key] = val
                    if dday_data[20] == 'Cooling':
                        for key, val in zip(DesignDay.COOLING_KEYS, dday_data[21:53]):
                            self._cooling_dict[key] = val
                    if dday_data[53] == 'Extremes':
                        for key, val in zip(DesignDay.EXTREME_KEYS, dday_data[54:70]):
                            self._extremes_dict[key] = val

                # parse typical and extreme periods into analysis periods.
                week_data = self._header[2].split(',')
                num_weeks = int(week_data[1]) if len(week_data) >= 2 else 0
                st_ind = 2
                for i in xrange(num_weeks):
                    week_dat = week_data[st_ind:st_ind + 4]
                    st_ind += 4
                    st = [int(num) for num in week_dat[2].split('/')]
                    end = [int(num) for num in week_dat[3].split('/')]
                    if len(st) == 3:
                        a_per = AnalysisPeriod(st[1], st[2], 0, end[1], end[2], 23)
                    elif len(st) == 2:
                        a_per = AnalysisPeriod(st[0], st[1], 0, end[0], end[1], 23)
                    if 'Max' in week_dat[0] and week_dat[1] == 'Extreme':
                        self._extreme_hot_weeks[week_dat[0]] = a_per
                    elif 'Min' in week_dat[0] and week_dat[1] == 'Extreme':
                        self._extreme_cold_weeks[week_dat[0]] = a_per
                    elif week_dat[1] == 'Typical':
                        self._typical_weeks[week_dat[0]] = a_per

                # parse the monthly ground temperatures in the header.
                grnd_data = self._header[3].strip().split(',')
                num_depths = int(grnd_data[1]) if len(grnd_data) >= 2 else 0
                st_ind = 2
                for i in xrange(num_depths):
                    header_meta = dict(self._metadata)  # copying the metadata dictionary
                    header_meta['depth'] = float(grnd_data[st_ind])
                    header_meta['soil conductivity'] = grnd_data[st_ind + 1]
                    header_meta['soil density'] = grnd_data[st_ind + 2]
                    header_meta['soil specific heat'] = grnd_data[st_ind + 3]
                    grnd_header = Header(temperature.GroundTemperature(), 'C',
                                         AnalysisPeriod(), header_meta)
                    grnd_vals = [float(x) for x in grnd_data[st_ind + 4: st_ind + 16]]
                    self._monthly_ground_temps[float(grnd_data[st_ind])] = \
                        MonthlyCollection(grnd_header, grnd_vals, list(xrange(12)))
                    st_ind += 16

                # parse leap year, daylight savings and comments.
                leap_dl_sav = self._header[4].strip().split(',')
                self._is_leap_year = True if leap_dl_sav[1] == 'Yes' else False
                self.daylight_savings_start = leap_dl_sav[2]
                self.daylight_savings_end = leap_dl_sav[3]
                comments_1 = self._header[5].strip().split(',')
                if len(comments_1) > 0:
                    self.comments_1 = ','.join(comments_1[1:])
                comments_2 = self._header[6].strip().split(',')
                if len(comments_2) > 0:
                    self.comments_2 = ','.join(comments_2[1:])

                self._is_header_loaded = True

            if import_header_only:
                return

            # read first line of data to overwrite the number of fields
            if original_header_load:
                for i in xrange(7):
                    epwin.readline()
            line = epwin.readline()
            self._num_of_fields = min(len(line.strip().split(',')), 35)

            # create an annual analysis period
            analysis_period = AnalysisPeriod(is_leap_year=self.is_leap_year)

            # create headers and an empty list for each field in epw file
            headers = []
            for field_number in xrange(self._num_of_fields):
                field = EPWFields.field_by_number(field_number)
                header = Header(data_type=field.name, unit=field.unit,
                                analysis_period=analysis_period,
                                metadata=dict(self._metadata))
                headers.append(header)
                self._data.append([])

            # collect hourly data
            while line:
                data = line.strip().split(',')

                for field_number in xrange(self._num_of_fields):
                    value_type = EPWFields.field_by_number(field_number).value_type
                    try:
                        value = value_type(data[field_number])
                    except ValueError as e:
                        # failed to convert the value for the specific TypeError
                        if value_type != int:
                            raise ValueError(e)
                        value = int(round(float(data[field_number])))

                    self._data[field_number].append(value)

                line = epwin.readline()

            # if the first value is at 1 AM, move last item to start position
            for field_number in xrange(self._num_of_fields):
                point_in_time = headers[field_number].data_type.point_in_time
                if point_in_time:
                    # move the last hour to first position
                    last_hour = self._data[field_number].pop()
                    self._data[field_number].insert(0, last_hour)

            # finally, build the data collection objects from the headers and data
            for i in xrange(self._num_of_fields):
                self._data[i] = HourlyContinuousCollection(headers[i], self._data[i])

            self._is_data_loaded = True

    @property
    def header(self):
        """A list of text representing the full header (the first 8 lines) of the EPW."""
        self._load_header_check()
        loc = self.location
        loc_str = 'LOCATION,{},{},{},{},{},{},{},{},{}\n'.format(
            loc.city, loc.state, loc.country, loc.source, loc.station_id, loc.latitude,
            loc.longitude, loc.time_zone, loc.elevation)
        winter_found = bool(self._heating_dict)
        summer_found = bool(self._cooling_dict)
        extreme_found = bool(self._extremes_dict)
        if winter_found and summer_found and extreme_found:
            des_str = 'DESIGN CONDITIONS,1,Climate Design Data 2009 ASHRAE Handbook,,'
            des_str = des_str + 'Heating,{},Cooling,{},Extremes,{}\n'.format(
                ','.join([self._heating_dict[key] for key in DesignDay.HEATING_KEYS]),
                ','.join([self._cooling_dict[key] for key in DesignDay.COOLING_KEYS]),
                ','.join([self._extremes_dict[key] for key in DesignDay.EXTREME_KEYS]))
        else:
            des_str = 'DESIGN CONDITIONS,0\n'
        weeks = []
        if bool(self.extreme_hot_weeks):
            for wk_name, a_per in self.extreme_hot_weeks.items():
                weeks.append(self._format_week(wk_name, 'Extreme', a_per))
        if bool(self.extreme_cold_weeks):
            for wk_name, a_per in self.extreme_cold_weeks.items():
                weeks.append(self._format_week(wk_name, 'Extreme', a_per))
        if bool(self.typical_weeks):
            for wk_name in sorted(self.typical_weeks.keys()):
                a_per = self.typical_weeks[wk_name]
                weeks.append(self._format_week(wk_name, 'Typical', a_per))
        week_str = 'TYPICAL/EXTREME PERIODS,{},{}\n'.format(len(weeks), ','.join(weeks))
        grnd_st = 'GROUND TEMPERATURES,{}'.format(len(self._monthly_ground_temps.keys()))
        for depth in sorted(self._monthly_ground_temps.keys()):
            grnd_st = grnd_st + ',{},{}'.format(
                depth, self._format_grndt(self._monthly_ground_temps[depth]))
        grnd_st = grnd_st + '\n'
        leap_yr = 'Yes' if self._is_leap_year else 'No'
        leap_str = 'HOLIDAYS/DAYLIGHT SAVINGS,{},{},{},0\n'.format(
            leap_yr, self.daylight_savings_start, self.daylight_savings_end)
        c_str1 = 'COMMENTS 1,{}\n'.format(self.comments_1)
        c_str2 = 'COMMENTS 2,{}\n'.format(self.comments_2)
        data_str = 'DATA PERIODS,1,1,Data,Sunday, 1/ 1,12/31\n'
        return [loc_str, des_str, week_str, grnd_st, leap_str, c_str1, c_str2, data_str]

    def _format_week(self, name, type, a_per):
        """Format an AnalysisPeriod into string for the EPW header."""
        return '{},{},{}/{},{}/{}'.format(name, type, a_per.st_month, a_per.st_day,
                                          a_per.end_month, a_per.end_day)

    def _format_grndt(self, data_c):
        """Format monthly ground data collection into string for the EPW header."""
        monthly_str = '{},{},{},{}'.format(
            data_c.header.metadata['soil conductivity'],
            data_c.header.metadata['soil density'],
            data_c.header.metadata['soil specific heat'],
            ','.join(['%.2f' % x for x in data_c.values]))
        return monthly_str

    def save(self, file_path):
        """Save epw object as an epw file.

        Args:
            file_path: A string representing the path to write the epw file to.
        """
        # load data if it's  not loaded convert to SI if it is in IP
        if not self.is_data_loaded:
            self._import_data()
        originally_ip = False
        if self.is_ip:
            self.convert_to_si()
            originally_ip = True

        # write the file
        lines = self.header
        try:
            # if the first value is at 1AM, move first item to end position
            for field in xrange(0, self._num_of_fields):
                point_in_time = self._data[field].header.data_type.point_in_time
                if point_in_time:
                    first_hour = self._data[field]._values.pop(0)
                    self._data[field]._values.append(first_hour)

            annual_a_per = AnalysisPeriod(is_leap_year=self.is_leap_year)
            for hour in xrange(0, len(annual_a_per.datetimes)):
                line = []
                for field in xrange(0, self._num_of_fields):
                    line.append(str(self._data[field]._values[hour]))
                lines.append(",".join(line) + "\n")
        except IndexError:
            # cleaning up
            length_error_msg = 'Data length is not for a full year and cannot be ' + \
                'saved as an EPW file.'
            raise ValueError(length_error_msg)
        else:
            file_data = ''.join(lines)
            write_to_file(file_path, file_data, True)
        finally:
            del(lines)
            # move last item to start position for fields on the hour
            for field in xrange(0, self._num_of_fields):
                point_in_time = self._data[field].header.data_type.point_in_time
                if point_in_time:
                    last_hour = self._data[field]._values.pop()
                    self._data[field]._values.insert(0, last_hour)

        if originally_ip:
            self.convert_to_ip()

        return file_path

    def convert_to_ip(self):
        """Convert all Data Collections of this EPW object to IP units.

        This is useful when one knows that all graphics produced from this
        EPW should be in Imperial units."""
        if not self.is_data_loaded:
            self._import_data()
        if not self.is_ip:
            for coll in self._data:
                coll.convert_to_ip()
        self._is_ip = True

    def convert_to_si(self):
        """Convert all Data Collections of this EPW object to SI units.

        This is useful when one needs to convert the EPW back to SI units
        from imperial units for processes like computing thermal comfort
        from EPW data."""
        if not self.is_data_loaded:
            self._import_data()
        if self.is_ip:
            for coll in self._data:
                coll.convert_to_si()
        self._is_ip = False

    def _get_data_by_field(self, field_number):
        """Return a data field by field number.

        This is a useful method to get the values for fields that Ladybug
        currently doesn't import by default. You can find list of fields by typing
        EPWFields.fields

        Args:
            field_number: a value between 0 to 34 for different available epw fields.

        Returns:
            An annual Ladybug list
        """
        if not self.is_data_loaded:
            self._import_data()

        # check input data
        if not 0 <= field_number < self._num_of_fields:
            raise ValueError("Field number should be between 0-%d" % self._num_of_fields)

        return self._data[field_number]

    def import_data_by_field(self, field_number):
        """Return an annual data collection for any field_number in epw file.

        This is a useful method to get the values for fields that Ladybug currently
        doesn't import by default. You can find list of fields by typing
        EPWFields.fields

        Args:
            field_number: A value between 0 to 34 for different available epw fields.

                *   0 Year
                *   1 Month
                *   2 Day
                *   3 Hour
                *   4 Minute
                *   5 -
                *   6 Dry Bulb Temperature
                *   7 Dew Point Temperature
                *   8 Relative Humidity
                *   9 Atmospheric Station Pressure
                *   10 Extraterrestrial Horizontal Radiation
                *   11 Extraterrestrial Direct Normal Radiation
                *   12 Horizontal Infrared Radiation Intensity
                *   13 Global Horizontal Radiation
                *   14 Direct Normal Radiation
                *   15 Diffuse Horizontal Radiation
                *   16 Global Horizontal Illuminance
                *   17 Direct Normal Illuminance
                *   18 Diffuse Horizontal Illuminance
                *   19 Zenith Luminance
                *   20 Wind Direction
                *   21 Wind Speed
                *   22 Total Sky Cover
                *   23 Opaque Sky Cover
                *   24 Visibility
                *   25 Ceiling Height
                *   26 Present Weather Observation
                *   27 Present Weather Codes
                *   28 Precipitable Water
                *   29 Aerosol Optical Depth
                *   30 Snow Depth
                *   31 Days Since Last Snowfall
                *   32 Albedo
                *   33 Liquid Precipitation Depth
                *   34 Liquid Precipitation Quantity

        Returns:
            An annual Ladybug list
        """
        return self._get_data_by_field(field_number)

    @property
    def years(self):
        """Return years as a Ladybug Data Collection."""
        return self._get_data_by_field(0)

    @property
    def dry_bulb_temperature(self):
        """Return annual Dry Bulb Temperature as a Ladybug Data Collection.

        This is the dry bulb temperature in C at the time indicated. Note that
        this is a full numeric field (i.e. 23.6) and not an integer representation
        with tenths. Valid values range from -70C to 70 C. Missing value for this
        field is 99.9.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(6)

    @property
    def dew_point_temperature(self):
        u"""Return annual Dew Point Temperature as a Ladybug Data Collection.

        This is the dew point temperature in C at the time indicated. Note that this is
        a full numeric field (i.e. 23.6) and not an integer representation with tenths.
        Valid values range from -70 C to 70 C. Missing value for this field is 99.9
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(7)

    @property
    def relative_humidity(self):
        u"""Return annual Relative Humidity as a Ladybug Data Collection.

        This is the Relative Humidity in percent at the time indicated. Valid values
        range from 0% to 110%. Missing value for this field is 999.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(8)

    @property
    def atmospheric_station_pressure(self):
        """Return annual Atmospheric Station Pressure as a Ladybug Data Collection.

        This is the station pressure in Pa at the time indicated. Valid values range
        from 31,000 to 120,000. (These values were chosen from the standard barometric
        pressure for all elevations of the World). Missing value for this field is 999999
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(9)

    @property
    def extraterrestrial_horizontal_radiation(self):
        """Return annual Extraterrestrial Horizontal Radiation as a Ladybug Data Collection.

        This is the Extraterrestrial Horizontal Radiation in Wh/m2. It is not currently
        used in EnergyPlus calculations. It should have a minimum value of 0; missing
        value for this field is 9999.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(10)

    @property
    def extraterrestrial_direct_normal_radiation(self):
        """Return annual Extraterrestrial Direct Normal Radiation as a Ladybug Data Collection.

        This is the Extraterrestrial Direct Normal Radiation in Wh/m2. (Amount of solar
        radiation in Wh/m2 received on a surface normal to the rays of the sun at the top
        of the atmosphere during the number of minutes preceding the time indicated).
        It is not currently used in EnergyPlus calculations. It should have a minimum
        value of 0; missing value for this field is 9999.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(11)

    @property
    def horizontal_infrared_radiation_intensity(self):
        """Return annual Horizontal Infrared Radiation Intensity as a Ladybug Data Collection.

        This is the Horizontal Infrared Radiation Intensity in W/m2. If it is missing,
        it is calculated from the Opaque Sky Cover field as shown in the following
        explanation. It should have a minimum value of 0; missing value for this field
        is 9999.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(12)

    @property
    def global_horizontal_radiation(self):
        """Return annual Global Horizontal Radiation as a Ladybug Data Collection.

        This is the Global Horizontal Radiation in Wh/m2. (Total amount of direct and
        diffuse solar radiation in Wh/m2 received on a horizontal surface during the
        number of minutes preceding the time indicated.) It is not currently used in
        EnergyPlus calculations. It should have a minimum value of 0; missing value
        for this field is 9999.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(13)

    @property
    def direct_normal_radiation(self):
        """Return annual Direct Normal Radiation as a Ladybug Data Collection.

        This is the Direct Normal Radiation in Wh/m2. (Amount of solar radiation in
        Wh/m2 received directly from the solar disk on a surface perpendicular to the
        sun's rays, during the number of minutes preceding the time indicated.) If the
        field is missing ( >= 9999) or invalid ( < 0), it is set to 0. Counts of such
        missing values are totaled and presented at the end of the runperiod.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(14)

    @property
    def diffuse_horizontal_radiation(self):
        """Return annual Diffuse Horizontal Radiation as a Ladybug Data Collection.

        This is the Diffuse Horizontal Radiation in Wh/m2. (Amount of solar radiation in
        Wh/m2 received from the sky (excluding the solar disk) on a horizontal surface
        during the number of minutes preceding the time indicated.) If the field is
        missing ( >= 9999) or invalid ( < 0), it is set to 0. Counts of such missing
        values are totaled and presented at the end of the runperiod
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(15)

    @property
    def global_horizontal_illuminance(self):
        """Return annual Global Horizontal Illuminance as a Ladybug Data Collection.

        This is the Global Horizontal Illuminance in lux. (Average total amount of
        direct and diffuse illuminance in hundreds of lux received on a horizontal
        surface during the number of minutes preceding the time indicated.) It is not
        currently used in EnergyPlus calculations. It should have a minimum value of 0;
        missing value for this field is 999999 and will be considered missing if greater
        than or equal to 999900.
        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(16)

    @property
    def direct_normal_illuminance(self):
        """Return annual Direct Normal Illuminance as a Ladybug Data Collection.

        This is the Direct Normal Illuminance in lux. (Average amount of illuminance in
        hundreds of lux received directly from the solar disk on a surface perpendicular
        to the sun's rays, during the number of minutes preceding the time indicated.)
        It is not currently used in EnergyPlus calculations. It should have a minimum
        value of 0; missing value for this field is 999999 and will be considered missing
        if greater than or equal to 999900.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(17)

    @property
    def diffuse_horizontal_illuminance(self):
        """Return annual Diffuse Horizontal Illuminance as a Ladybug Data Collection.

        This is the Diffuse Horizontal Illuminance in lux. (Average amount of illuminance
        in hundreds of lux received from the sky (excluding the solar disk) on a
        horizontal surface during the number of minutes preceding the time indicated.)
        It is not currently used in EnergyPlus calculations. It should have a minimum
        value of 0; missing value for this field is 999999 and will be considered missing
        if greater than or equal to 999900.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(18)

    @property
    def zenith_luminance(self):
        """Return annual Zenith Luminance as a Ladybug Data Collection.

        This is the Zenith Illuminance in Cd/m2. (Average amount of luminance at
        the sky's zenith in tens of Cd/m2 during the number of minutes preceding
        the time indicated.) It is not currently used in EnergyPlus calculations.
        It should have a minimum value of 0; missing value for this field is 9999.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(19)

    @property
    def wind_direction(self):
        """Return annual Wind Direction as a Ladybug Data Collection.

        This is the Wind Direction in degrees where the convention is that North=0.0,
        East=90.0, South=180.0, West=270.0. (Wind direction in degrees at the time
        indicated. If calm, direction equals zero.) Values can range from 0 to 360.
        Missing value is 999.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(20)

    @property
    def wind_speed(self):
        """Return annual Wind Speed as a Ladybug Data Collection.

        This is the wind speed in m/sec. (Wind speed at time indicated.) Values can
        range from 0 to 40. Missing value is 999.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(21)

    @property
    def total_sky_cover(self):
        """Return annual Total Sky Cover as a Ladybug Data Collection.

        This is the value for total sky cover (tenths of coverage). (i.e. 1 is 1/10
        covered. 10 is total coverage). (Amount of sky dome in tenths covered by clouds
        or obscuring phenomena at the hour indicated at the time indicated.) Minimum
        value is 0; maximum value is 10; missing value is 99.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(22)

    @property
    def opaque_sky_cover(self):
        """Return annual Opaque Sky Cover as a Ladybug Data Collection.

        This is the value for opaque sky cover (tenths of coverage). (i.e. 1 is 1/10
        covered. 10 is total coverage). (Amount of sky dome in tenths covered by
        clouds or obscuring phenomena that prevent observing the sky or higher cloud
        layers at the time indicated.) This is not used unless the field for Horizontal
        Infrared Radiation Intensity is missing and then it is used to calculate
        Horizontal Infrared Radiation Intensity. Minimum value is 0; maximum value is
        10; missing value is 99.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(23)

    @property
    def visibility(self):
        """Return annual Visibility as a Ladybug Data Collection.

        This is the value for visibility in km. (Horizontal visibility at the time
        indicated.) It is not currently used in EnergyPlus calculations. Missing
        value is 9999.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(24)

    @property
    def ceiling_height(self):
        """Return annual Ceiling Height as a Ladybug Data Collection.

        This is the value for ceiling height in m. (77777 is unlimited ceiling height.
        88888 is cirroform ceiling.) It is not currently used in EnergyPlus calculations.
        Missing value is 99999

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(25)

    @property
    def present_weather_observation(self):
        """Return annual Present Weather Observation as a Ladybug Data Collection.

        If the value of the field is 0, then the observed weather codes are taken from
        the following field. If the value of the field is 9, then "missing" weather is
        assumed. Since the primary use of these fields (Present Weather Observation and
        Present Weather Codes) is for rain/wet surfaces, a missing observation field or
        a missing weather code implies no rain.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(26)

    @property
    def present_weather_codes(self):
        """Return annual Present Weather Codes as a Ladybug Data Collection.

        The present weather codes field is assumed to follow the TMY2 conventions for
        this field. Note that though this field may be represented as numeric (e.g. in
        the CSV format), it is really a text field of 9 single digits. This convention
        along with values for each "column" (left to right) is presented in Table 16.
        Note that some formats (e.g. TMY) does not follow this convention - as much as
        possible, the present weather codes are converted to this convention during
        WeatherConverter processing. Also note that the most important fields are those
        representing liquid precipitation - where the surfaces of the building would be
        wet. EnergyPlus uses "Snow Depth" to determine if snow is on the ground.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(27)

    @property
    def precipitable_water(self):
        """Return annual Precipitable Water as a Ladybug Data Collection.

        This is the value for Precipitable Water in mm. (This is not rain - rain is
        inferred from the PresWeathObs field but a better result is from the Liquid
        Precipitation Depth field). It is not currently used in EnergyPlus calculations
        (primarily due to the unreliability of the reporting of this value). Missing
        value is 999.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(28)

    @property
    def aerosol_optical_depth(self):
        """Return annual Aerosol Optical Depth as a Ladybug Data Collection.

        This is the value for Aerosol Optical Depth in thousandths. It is not currently
        used in EnergyPlus calculations. Missing value is .999.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(29)

    @property
    def snow_depth(self):
        """Return annual Snow Depth as a Ladybug Data Collection.

        This is the value for Snow Depth in cm. This field is used to tell when snow
        is on the ground and, thus, the ground reflectance may change. Missing value
        is 999.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(30)

    @property
    def days_since_last_snowfall(self):
        """Return annual Days Since Last Snow Fall as a Ladybug Data Collection.

        This is the value for Days Since Last Snowfall. It is not currently used in
        EnergyPlus calculations. Missing value is 99.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(31)

    @property
    def albedo(self):
        """Return annual Albedo values as a Ladybug Data Collection.

        The ratio (unitless) of reflected solar irradiance to global horizontal
        irradiance. It is not currently used in EnergyPlus.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(32)

    @property
    def liquid_precipitation_depth(self):
        """Return annual liquid precipitation depth as a Ladybug Data Collection.

        The amount of liquid precipitation (mm) observed at the indicated time for the
        period indicated in the liquid precipitation quantity field. If this value is
        not missing, then it is used and overrides the "precipitation" flag as rainfall.
        Conversely, if the precipitation flag shows rain and this field is missing or
        zero, it is set to 1.5 (mm).

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs\
/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(33)

    @property
    def liquid_precipitation_quantity(self):
        """Return annual Liquid Precipitation Quantity as a Ladybug Data Collection.

        The period of accumulation (hr) for the liquid precipitation depth field.
        It is not currently used in EnergyPlus.

        Read more at: https://energyplus.net/sites/all/modules/custom/nrel_custom/\
pdfs/pdfs_v8.4.0/AuxiliaryPrograms.pdf (Chapter 2.9.1)
        """
        return self._get_data_by_field(34)

    @property
    def sky_temperature(self):
        """Return annual Sky Temperature as a Ladybug Data Collection.

        This value in degrees Celsius is derived from the Horizontal Infrared
        Radiation Intensity in Wh/m2. It represents the long wave radiant
        temperature of the sky
        Read more at: https://bigladdersoftware.com/epx/docs/8-9/engineering-reference\
/climate-calculations.html#energyplus-sky-temperature-calculation
        """
        # create sky temperature header
        sky_temp_header = Header(data_type=temperature.SkyTemperature(), unit='C',
                                 analysis_period=AnalysisPeriod(),
                                 metadata=self._metadata)

        # calculate sy temperature for each hour
        horiz_ir = self._get_data_by_field(12).values
        sky_temp_data = [calc_sky_temperature(hir) for hir in horiz_ir]
        return HourlyContinuousCollection(sky_temp_header, sky_temp_data)

    def _get_wea_header(self):
        return "place %s\n" % self.location.city + \
            "latitude %.2f\n" % self.location.latitude + \
            "longitude %.2f\n" % -self.location.longitude + \
            "time_zone %d\n" % (-self.location.time_zone * 15) + \
            "site_elevation %.1f\n" % self.location.elevation + \
            "weather_data_file_unit 1\n"

    def to_wea(self, file_path, hoys=None):
        """Write an wea file from the epw file.

        WEA carries radiation values from epw. Gendaymtx uses these values to
        generate the sky. For an annual analysis it is identical to using epw2wea.

        Args:
            file_path: Full file path for output file.
            hoys: List of hours of the year. Default is 0-8759.
        """
        hoys = hoys or xrange(len(self.direct_normal_radiation.datetimes))
        if not file_path.lower().endswith('.wea'):
            file_path += '.wea'

        originally_ip = False
        if self.is_ip:
            self.convert_to_si()
            originally_ip = True

        # write header
        lines = [self._get_wea_header()]
        # write values
        datetimes = self.direct_normal_radiation.datetimes
        for hoy in hoys:
            dir_rad = self.direct_normal_radiation[hoy]
            dif_rad = self.diffuse_horizontal_radiation[hoy]
            line = "%d %d %.3f %d %d\n" \
                % (datetimes[hoy].month,
                   datetimes[hoy].day,
                   datetimes[hoy].hour + 0.5,
                   dir_rad, dif_rad)
            lines.append(line)

        file_data = ''.join(lines)
        write_to_file(file_path, file_data, True)

        if originally_ip:
            self.convert_to_ip()

        return file_path

    def to_dict(self):
        """Convert the EPW to a dictionary."""
        # load data if it's not loaded
        if not self.is_data_loaded:
            self._import_data()

        def dictify_dict(base_dict):
            new_dict = {}
            for key, val in base_dict.items():
                new_dict[key] = val.to_dict()
            return new_dict
        hot_wks = dictify_dict(self.extreme_hot_weeks)
        cold_wks = dictify_dict(self.extreme_cold_weeks)
        typ_wks = dictify_dict(self.typical_weeks)
        grnd_temps = dictify_dict(self.monthly_ground_temperature)
        return {
            'location': self.location.to_dict(),
            'data_collections': [dc.to_dict() for dc in self._data],
            'metadata': self.metadata,
            'heating_dict': self.heating_design_condition_dictionary,
            'cooling_dict': self.cooling_design_condition_dictionary,
            'extremes_dict': self.extreme_design_condition_dictionary,
            'extreme_hot_weeks': hot_wks,
            'extreme_cold_weeks': cold_wks,
            'typical_weeks': typ_wks,
            "monthly_ground_temps": grnd_temps,
            "is_ip": self._is_ip,
            "is_leap_year": self.is_leap_year,
            "daylight_savings_start": self.daylight_savings_start,
            "daylight_savings_end": self.daylight_savings_end,
            "comments_1": self.comments_1,
            "comments_2": self.comments_2,
            "type": 'EPW'
        }

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """epw file representation."""
        return "EPW file Data for [%s]" % self.location.city


class EPWFields(object):
    """EPW weather file fields.

    Read more at https://energyplus.net/sites/all/modules/custom/nrel_custom/pdfs/\
pdfs_v8.4.0/AuxiliaryPrograms.pdf
    (Chapter 2.9.1)
    """

    _fields = {
        0: {'name': generic.GenericType('Year', 'yr'),
            'type': int,
            'unit': 'yr'
            },

        1: {'name': generic.GenericType('Month', 'mon'),
            'type': int,
            'unit': 'mon'
            },

        2: {'name': generic.GenericType('Day', 'day'),
            'type': int,
            'unit': 'day'
            },

        3: {'name': generic.GenericType('Hour', 'hr'),
            'type': int,
            'unit': 'hr'
            },

        4: {'name': generic.GenericType('Minute', 'min'),
            'type': int,
            'unit': 'min'
            },

        5: {'name': generic.GenericType('Uncertainty Flags', 'flag'),
            'type': str,
            'unit': 'flag'
            },

        6: {'name': temperature.DryBulbTemperature(),
            'type': float,
            'unit': 'C',
            'min': -70,
            'max': 70,
            'missing': 99.9
            },

        7: {'name': temperature.DewPointTemperature(),
            'type': float,
            'unit': 'C',
            'min': -70,
            'max': 70,
            'missing': 99.9
            },

        8: {'name': fraction.RelativeHumidity(),
            'type': int,
            'unit': '%',
            'missing': 999,
            'min': 0,
            'max': 110
            },

        9: {'name': pressure.AtmosphericStationPressure(),
            'type': int,
            'unit': 'Pa',
            'missing': 999999,
            'min': 31000,
            'max': 120000
            },

        10: {'name': energyintensity.ExtraterrestrialHorizontalRadiation(),
             'type': int,
             'unit': 'Wh/m2',
             'missing': 9999,
             'min': 0
             },

        11: {'name': energyintensity.ExtraterrestrialDirectNormalRadiation(),
             'type': int,
             'unit': 'Wh/m2',
             'missing': 9999,
             'min': 0
             },

        12: {'name': energyflux.HorizontalInfraredRadiationIntensity(),
             'type': int,
             'unit': 'W/m2',
             'missing': 9999,
             'min': 0
             },

        13: {'name': energyintensity.GlobalHorizontalRadiation(),
             'type': int,
             'unit': 'Wh/m2',
             'missing': 9999,
             'min': 0
             },

        14: {'name': energyintensity.DirectNormalRadiation(),
             'type': int,
             'unit': 'Wh/m2',
             'missing': 9999,
             'min': 0
             },

        15: {'name': energyintensity.DiffuseHorizontalRadiation(),
             'type': int,
             'unit': 'Wh/m2',
             'missing': 9999,
             'min': 0
             },

        16: {'name': illuminance.GlobalHorizontalIlluminance(),
             'type': int,
             'unit': 'lux',
             'missing': 999999,  # will be missing if >= 999900
             'min': 0
             },

        17: {'name': illuminance.DirectNormalIlluminance(),
             'type': int,
             'unit': 'lux',
             'missing': 999999,  # will be missing if >= 999900
             'min': 0
             },

        18: {'name': illuminance.DiffuseHorizontalIlluminance(),
             'type': int,
             'unit': 'lux',
             'missing': 999999,  # will be missing if >= 999900
             'min': 0
             },

        19: {'name': luminance.ZenithLuminance(),
             'type': int,
             'unit': 'cd/m2',
             'missing': 9999,  # will be missing if >= 9999
             'min': 0
             },

        20: {'name': angle.WindDirection(),
             'type': int,
             'unit': 'degrees',
             'missing': 999,
             'min': 0,
             'max': 360
             },

        21: {'name': speed.WindSpeed(),
             'type': float,
             'unit': 'm/s',
             'missing': 999,
             'min': 0,
             'max': 40
             },

        22: {'name': fraction.TotalSkyCover(),  # used if Horizontal IR is missing
             'type': int,
             'unit': 'tenths',
             'missing': 99,
             'min': 0,
             'max': 10
             },

        23: {'name': fraction.OpaqueSkyCover(),  # used if Horizontal IR is missing
             'type': int,
             'unit': 'tenths',
             'missing': 99
             },

        24: {'name': distance.Visibility(),
             'type': float,
             'unit': 'km',
             'missing': 9999
             },

        25: {'name': distance.CeilingHeight(),
             'type': int,
             'unit': 'm',
             'missing': 99999
             },

        26: {'name': generic.GenericType(name='Present Weather Observation',
                                         unit='observation'),
             'type': int,
             'unit': 'observation',
             'missing': 9
             },

        27: {'name': generic.GenericType(name='Present Weather Codes', unit='codes'),
             'type': int,
             'unit': 'codes',
             'missing': 999999999
             },

        28: {'name': distance.PrecipitableWater(),
             'type': int,
             'unit': 'mm',
             'missing': 999
             },

        29: {'name': fraction.AerosolOpticalDepth(),
             'type': float,
             'unit': 'fraction',
             'missing': 999
             },

        30: {'name': distance.SnowDepth(),
             'type': int,
             'unit': 'cm',
             'missing': 999
             },

        31: {'name': generic.GenericType(name='Days Since Last Snowfall', unit='day'),
             'type': int,
             'unit': 'day',
             'missing': 99
             },

        32: {'name': fraction.Albedo(),
             'type': float,
             'unit': 'fraction',
             'missing': 999
             },

        33: {'name': distance.LiquidPrecipitationDepth(),
             'type': float,
             'unit': 'mm',
             'missing': 999
             },

        34: {'name': fraction.LiquidPrecipitationQuantity(),
             'type': float,
             'unit': 'fraction',
             'missing': 99
             }
    }

    @classmethod
    def field_by_number(cls, field_number):
        """Return an EPWField based on field number.

        *   0 Year
        *   1 Month
        *   2 Day
        *   3 Hour
        *   4 Minute
        *   5 -
        *   6 Dry Bulb Temperature
        *   7 Dew Point Temperature
        *   8 Relative Humidity
        *   9 Atmospheric Station Pressure
        *   10 Extraterrestrial Horizontal Radiation
        *   11 Extraterrestrial Direct Normal Radiation
        *   12 Horizontal Infrared Radiation Intensity
        *   13 Global Horizontal Radiation
        *   14 Direct Normal Radiation
        *   15 Diffuse Horizontal Radiation
        *   16 Global Horizontal Illuminance
        *   17 Direct Normal Illuminance
        *   18 Diffuse Horizontal Illuminance
        *   19 Zenith Luminance
        *   20 Wind Direction
        *   21 Wind Speed
        *   22 Total Sky Cover
        *   23 Opaque Sky Cover
        *   24 Visibility
        *   25 Ceiling Height
        *   26 Present Weather Observation
        *   27 Present Weather Codes
        *   28 Precipitable Water
        *   29 Aerosol Optical Depth
        *   30 Snow Depth
        *   31 Days Since Last Snowfall
        *   32 Albedo
        *   33 Liquid Precipitation Depth
        *   34 Liquid Precipitation Quantity
        """
        return EPWField(cls._fields[field_number])

    def __repr__(self):
        """EPW fields representation."""
        fields = (
            '{}: {}'.format(key, value['name'])
            for key, value in self._fields.items()
        )

        return '\n'.join(fields)


class EPWField(object):
    """An EPW field.

    Attributes:
        name: Name of the field.
        value_type: field value type (e.g. int, float, str)
        unit: Field unit.
        missing: Missing value for the data type in EPW files.
    """
    __slots__ = ('name', 'value_type', 'unit', 'missing')

    def __init__(self, field_dict):
        self.name = field_dict['name']
        self.value_type = field_dict['type']
        if 'unit' in field_dict:
            self.unit = field_dict['unit']
        else:
            self.unit = None
        if 'missing' in field_dict:
            self.missing = field_dict['missing']
        else:
            self.missing = None



# coding=utf-8
from ladybug.epw import EPW
from ladybug.datacollection import HourlyContinuousCollection, MonthlyCollection
from ladybug.designday import DesignDay
from ladybug.analysisperiod import AnalysisPeriod

import os
import pytest


def test_import_epw():
    """Test import standard epw."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    abs_path = os.path.abspath(relative_path)
    epw_rel = EPW(relative_path)
    epw = EPW(abs_path)

    assert epw_rel.file_path == os.path.normpath(relative_path)
    assert epw_rel.location.city == 'Chicago Ohare Intl Ap'
    assert epw.file_path == abs_path
    assert epw.location.city == 'Chicago Ohare Intl Ap'
    # Check that calling location getter only retrieves location
    assert not epw.is_data_loaded
    dbt = epw.dry_bulb_temperature
    skyt = epw.sky_temperature  # test sky temperature calculation
    assert epw.is_data_loaded
    assert len(dbt) == 8760
    assert len(skyt) == 8760


def test_import_tokyo_epw():
    """Test import standard epw from another location."""
    path = './tests/fixtures/epw/tokyo.epw'
    epw = EPW(path)
    assert not epw.is_header_loaded
    assert epw.location.city == 'Tokyo'
    assert epw.is_header_loaded
    assert not epw.is_data_loaded
    dbt = epw.dry_bulb_temperature
    assert epw.is_data_loaded
    assert len(dbt) == 8760


def test_epw_from_missing_values():
    """Test import custom epw with wrong types."""
    epw = EPW.from_missing_values()
    assert epw.is_header_loaded
    assert epw.is_data_loaded
    assert len(epw.dry_bulb_temperature) == 8760
    assert list(epw.dry_bulb_temperature.values) == [99.9] * 8760


def test_dict_methods():
    """Test JSON serialization methods"""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)

    epw_dict = epw.to_dict()
    rebuilt_epw = EPW.from_dict(epw_dict)
    assert epw_dict == rebuilt_epw.to_dict()


def test_invalid_epw():
    """Test the import of incorrect file type and a non-existent epw file."""
    path = './tests/fixtures/epw/non-exitent.epw'
    with pytest.raises(Exception):
        epw = EPW(path)
        epw.location

    path = './tests/fixtures/stat/chicago.stat'
    with pytest.raises(Exception):
        epw = EPW(path)
        epw.location


def test_import_data():
    """Test the imported data properties."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    assert isinstance(epw.years, HourlyContinuousCollection)
    assert isinstance(epw.dry_bulb_temperature, HourlyContinuousCollection)
    assert isinstance(epw.dew_point_temperature, HourlyContinuousCollection)
    assert isinstance(epw.relative_humidity, HourlyContinuousCollection)
    assert isinstance(epw.atmospheric_station_pressure, HourlyContinuousCollection)
    assert isinstance(epw.extraterrestrial_horizontal_radiation, HourlyContinuousCollection)
    assert isinstance(epw.extraterrestrial_direct_normal_radiation, HourlyContinuousCollection)
    assert isinstance(epw.horizontal_infrared_radiation_intensity, HourlyContinuousCollection)
    assert isinstance(epw.global_horizontal_radiation, HourlyContinuousCollection)
    assert isinstance(epw.direct_normal_radiation, HourlyContinuousCollection)
    assert isinstance(epw.diffuse_horizontal_radiation, HourlyContinuousCollection)
    assert isinstance(epw.global_horizontal_illuminance, HourlyContinuousCollection)
    assert isinstance(epw.direct_normal_illuminance, HourlyContinuousCollection)
    assert isinstance(epw.diffuse_horizontal_illuminance, HourlyContinuousCollection)
    assert isinstance(epw.zenith_luminance, HourlyContinuousCollection)
    assert isinstance(epw.wind_direction, HourlyContinuousCollection)
    assert isinstance(epw.wind_speed, HourlyContinuousCollection)
    assert isinstance(epw.total_sky_cover, HourlyContinuousCollection)
    assert isinstance(epw.opaque_sky_cover, HourlyContinuousCollection)
    assert isinstance(epw.visibility, HourlyContinuousCollection)
    assert isinstance(epw.ceiling_height, HourlyContinuousCollection)
    assert isinstance(epw.present_weather_observation, HourlyContinuousCollection)
    assert isinstance(epw.present_weather_codes, HourlyContinuousCollection)
    assert isinstance(epw.precipitable_water, HourlyContinuousCollection)
    assert isinstance(epw.aerosol_optical_depth, HourlyContinuousCollection)
    assert isinstance(epw.snow_depth, HourlyContinuousCollection)
    assert isinstance(epw.days_since_last_snowfall, HourlyContinuousCollection)
    assert isinstance(epw.albedo, HourlyContinuousCollection)
    assert isinstance(epw.liquid_precipitation_depth, HourlyContinuousCollection)
    assert isinstance(epw.liquid_precipitation_quantity, HourlyContinuousCollection)
    assert isinstance(epw.sky_temperature, HourlyContinuousCollection)


def test_convert_to_ip():
    """Test the method that converts the data to IP units."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)

    assert epw.dry_bulb_temperature.header.unit == 'C'
    assert epw.dry_bulb_temperature.values[0] == -6.1
    epw.convert_to_ip()
    assert epw.dry_bulb_temperature.header.unit == 'F'
    assert epw.dry_bulb_temperature.values[0] == pytest.approx(21.02, rel=1e-2)
    epw.convert_to_si()
    assert epw.dry_bulb_temperature.header.unit == 'C'
    assert epw.dry_bulb_temperature.values[0] == pytest.approx(-6.1, rel=1e-5)


def test_set_data():
    """Test the ability to set the data of any of the epw hourly data."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    epw.dry_bulb_temperature[12] = 20
    assert epw.dry_bulb_temperature[12] == 20
    epw.dry_bulb_temperature.values = list(range(8760))
    assert epw.dry_bulb_temperature.values == tuple(range(8760))

    # Test if the set data is not annual
    with pytest.raises(Exception):
        epw.dry_bulb_temperature = list(range(365))


def test_import_design_conditions():
    """Test the functions that import design conditions."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    assert isinstance(epw.heating_design_condition_dictionary, dict)
    assert len(epw.heating_design_condition_dictionary.keys()) == 15
    assert isinstance(epw.cooling_design_condition_dictionary, dict)
    assert len(epw.cooling_design_condition_dictionary.keys()) == 32
    assert isinstance(epw.extreme_design_condition_dictionary, dict)
    assert len(epw.extreme_design_condition_dictionary.keys()) == 16


def test_set_design_conditions():
    """Test the functions that set design conditions."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)

    heat_dict = dict(epw.heating_design_condition_dictionary)
    heat_dict['DB996'] = -25
    epw.heating_design_condition_dictionary = heat_dict
    assert epw.heating_design_condition_dictionary['DB996'] == -25

    # Check for when the dictionary has a missing key
    wrong_dict = dict(heat_dict)
    del wrong_dict['DB996']
    with pytest.raises(Exception):
        epw.heating_design_condition_dictionary = wrong_dict

    # Check for when the wrong type is assigned
    heat_list = list(epw.heating_design_condition_dictionary.keys())
    with pytest.raises(Exception):
        epw.heating_design_condition_dictionary = heat_list

    cool_dict = dict(epw.cooling_design_condition_dictionary)
    cool_dict['DB004'] = 40
    epw.cooling_design_condition_dictionary = cool_dict
    assert epw.cooling_design_condition_dictionary['DB004'] == 40

    extremes_dict = dict(epw.extreme_design_condition_dictionary)
    extremes_dict['WS010'] = 20
    epw.extreme_design_condition_dictionary = extremes_dict
    assert epw.extreme_design_condition_dictionary['WS010'] == 20


def test_import_design_days():
    """Test the functions that import design days."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    assert isinstance(epw.annual_heating_design_day_996, DesignDay)
    assert epw.annual_heating_design_day_996.dry_bulb_condition.dry_bulb_max == -20.0
    assert isinstance(epw.annual_heating_design_day_990, DesignDay)
    assert epw.annual_heating_design_day_990.dry_bulb_condition.dry_bulb_max == -16.6
    assert isinstance(epw.annual_cooling_design_day_004, DesignDay)
    assert epw.annual_cooling_design_day_004.dry_bulb_condition.dry_bulb_max == 33.3
    assert isinstance(epw.annual_cooling_design_day_010, DesignDay)
    assert epw.annual_cooling_design_day_010.dry_bulb_condition.dry_bulb_max == 31.6


def test_import_extreme_weeks():
    """Test the functions that import the extreme weeks."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    ext_cold = list(epw.extreme_cold_weeks.values())[0]
    ext_hot = list(epw.extreme_hot_weeks.values())[0]
    assert isinstance(ext_cold, AnalysisPeriod)
    assert len(ext_cold.doys_int) == 7
    assert (ext_cold.st_month, ext_cold.st_day, ext_cold.end_month,
            ext_cold.end_day) == (1, 27, 2, 2)
    assert isinstance(ext_hot, AnalysisPeriod)
    assert len(ext_hot.doys_int) == 7
    assert (ext_hot.st_month, ext_hot.st_day, ext_hot.end_month,
            ext_hot.end_day) == (7, 13, 7, 19)


def test_import_typical_weeks():
    """Test the functions that import the typical weeks."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    typ_weeks = list(epw.typical_weeks.values())
    assert len(typ_weeks) == 4
    for week in typ_weeks:
        assert isinstance(week, AnalysisPeriod)
        assert len(week.doys_int) == 7


def test_set_extreme_typical_weeks():
    """Test the functions that set the extreme  and typical weeks."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    a_per_cold = AnalysisPeriod(1, 1, 0, 1, 7, 23)
    a_per_hot = AnalysisPeriod(7, 1, 0, 7, 7, 23)
    a_per_typ = AnalysisPeriod(5, 1, 0, 5, 7, 23)
    epw.extreme_cold_weeks = {'Extreme Cold Week': a_per_cold}
    epw.extreme_hot_weeks = {'Extreme Hot Week': a_per_hot}
    epw.typical_weeks = {'Typical Week': a_per_typ}
    assert list(epw.extreme_cold_weeks.values())[0] == a_per_cold
    assert list(epw.extreme_hot_weeks.values())[0] == a_per_hot
    assert list(epw.typical_weeks.values())[0] == a_per_typ

    # Test one someone sets an analysis_period longer than a week.
    a_per_wrong = AnalysisPeriod(1, 1, 0, 1, 6, 23)
    with pytest.raises(Exception):
        epw.extreme_cold_weeks = {'Extreme Cold Week': a_per_wrong}

    # Test when someone sets the wrong type of data
    with pytest.raises(Exception):
        epw.extreme_cold_weeks = a_per_cold


def test_import_ground_temperatures():
    """Test the functions that import ground temperature."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    assert len(epw.monthly_ground_temperature.keys()) == 3
    assert tuple(epw.monthly_ground_temperature.keys()) == (0.5, 2.0, 4.0)
    assert isinstance(epw.monthly_ground_temperature[0.5], MonthlyCollection)
    assert epw.monthly_ground_temperature[0.5].values == \
        (-1.89, -3.06, -0.99, 2.23, 10.68, 17.2,
         21.6, 22.94, 20.66, 15.6, 8.83, 2.56)
    assert epw.monthly_ground_temperature[2].values == \
        (2.39, 0.31, 0.74, 2.45, 8.1, 13.21,
         17.3, 19.5, 19.03, 16.16, 11.5, 6.56)
    assert epw.monthly_ground_temperature[4].values == \
        (5.93, 3.8, 3.34, 3.98, 7.18, 10.62,
         13.78, 15.98, 16.49, 15.25, 12.51, 9.17)


def test_set_ground_temperatures():
    """Test the functions that set ground temperature."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    grnd_dict = dict(epw.monthly_ground_temperature)
    grnd_dict[0.5].values = list(range(12))
    epw.monthly_ground_temperature = grnd_dict
    assert epw.monthly_ground_temperature[0.5].values == tuple(range(12))

    # test when the type is not a monthly collection.
    grnd_dict = dict(epw.monthly_ground_temperature)
    grnd_dict[0.5] = list(range(12))
    with pytest.raises(Exception):
        epw.monthly_ground_temperature = grnd_dict


def test_epw_header():
    """Check that the process of parsing the EPW header hasn't changed it."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    for i in range(len(epw.header)):
        line1, line2 = epw.header[i], epw._header[i]
        if i in (0, 1, 4, 5, 6, 7):
            # These lines should match exactly
            assert line1.rstrip() == line2.rstrip()
        elif i in (2, 3):
            # The order of data in these lines can change and  spaces can get deleted
            assert len(line1.split(',')) == len(line2.split(','))


def test_save_epw():
    """Test save epw_rel."""
    path = './tests/fixtures/epw/tokyo.epw'
    epw = EPW(path)

    modified_path = './tests/fixtures/epw/tokyo_modified.epw'
    epw.save(modified_path)
    assert os.path.isfile(modified_path)
    assert os.stat(modified_path).st_size > 1
    os.remove(modified_path)


def test_save_epw_from_missing_values():
    """Test import custom epw with wrong types."""
    epw = EPW.from_missing_values()
    file_path = './tests/fixtures/epw/missing.epw'
    epw.save(file_path)
    assert os.path.isfile(file_path)
    assert os.stat(file_path).st_size > 1
    os.remove(file_path)


def test_save_converted_epw():
    """Test that the saved EPW always has SI units."""
    relative_path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(relative_path)
    epw.convert_to_ip()
    modified_path = './tests/fixtures/epw/chicago_modified.epw'
    epw.save(modified_path)
    assert epw.dry_bulb_temperature.header.unit == 'F'
    assert epw.dry_bulb_temperature.values[0] == pytest.approx(21.02, rel=1e-2)

    new_epw = EPW(modified_path)
    assert new_epw.dry_bulb_temperature.header.unit == 'C'
    assert new_epw.dry_bulb_temperature.values[0] == pytest.approx(-6.1, rel=1e-5)
    os.remove(modified_path)


def test_save_wea():
    """Test save wea_rel."""
    path = './tests/fixtures/epw/chicago.epw'
    epw = EPW(path)
    wea_path = './tests/fixtures/wea/chicago_epw.wea'
    epw.to_wea(wea_path)
    assert os.path.isfile(wea_path)
    assert os.stat(wea_path).st_size > 1

    # check the order of the data in the file
    with open(wea_path) as wea_f:
        line = wea_f.readlines()
        assert float(line[6].split(' ')[-2]) == epw.direct_normal_radiation[0]
        assert float(line[6].split(' ')[-1]) == epw.diffuse_horizontal_radiation[0]
        assert float(line[17].split(' ')[-2]) == epw.direct_normal_radiation[11]
        assert float(line[17].split(' ')[-1]) == epw.diffuse_horizontal_radiation[11]

    os.remove(wea_path)



# coding=utf-8
from __future__ import division

import codecs
import os
import platform
import re

from .analysisperiod import AnalysisPeriod
from .designday import ASHRAEClearSky
from .designday import ASHRAETau
from .designday import DesignDay
from .designday import DryBulbCondition
from .designday import HumidityCondition
from .designday import WindCondition
from .dt import Date
from .location import Location

try:
    from itertools import izip as zip  # python 2
except ImportError:
    xrange = range  # python 3


class STAT(object):
    """Import data from a local .stat file.

    Args:
        file_path: Address to a local .stat file.

    Properties:
        * location
        * ashrae_climate_zone
        * koppen_climate_zone
        * extreme_cold_week
        * extreme_hot_week
        * typical_winter_week
        * typical_spring_week
        * typical_summer_week
        * typical_autumn_week
        * other_typical_weeks
        * annual_heating_design_day_996
        * annual_heating_design_day_990
        * annual_cooling_design_day_004
        * annual_cooling_design_day_010
        * monthly_cooling_design_days_100
        * monthly_cooling_design_days_050
        * monthly_cooling_design_days_020
        * monthly_cooling_design_days_004
        * monthly_db_temp_050
        * monthly_wb_temp_050
        * monthly_db_temp_range_050
        * monthly_wb_temp_range_050
        * monthly_found
        * standard_pressure_at_elev
        * monthly_wind_conditions
        * monthly_ws_avg
        * monthly_wind_dirs
        * monthly_clear_sky_conditions
        * monthly_tau_beam
        * monthly_tau_diffuse
        * file_path
    """
    # categories used for parsing text
    _months = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul',
               'Aug', 'Sep', 'Oct', 'Nov', 'Dec')
    _wind_dirs = (0, 45, 90, 135, 180, 225, 270, 315)
    _wind_dir_names = ('North', 'NorthEast', 'East', 'SouthEast', 'South',
                       'SouthWest', 'West', 'NorthWest')
    # compiled strings for identifying data in the file
    _coord_pattern1 = re.compile(r"{([NSEW])(\s*\d*)deg(\s*\d*)")
    _coord_pattern2 = re.compile(r"{([NSEW])(\s*\d*) (\s*\d*)")
    _elev_pattern1 = re.compile(r"Elevation\s*[-]*\s*(\d*)m\s*(\S*)")
    _elev_pattern2 = re.compile(r"Elevation\s*[-]*\s*(\d*)\s*m\s*(\S*)")
    _timez_pattern = re.compile(r"{GMT\s*(\S*)\s*Hours}")
    _press_pattern = re.compile(r"Elevation\s*[-]*\s*(\d*)")
    _ashraecz_pattern = re.compile(r'Climate type\s"(\S*)"\s\(A')
    _koppencz_pattern = re.compile(r'Climate type\s"(\S*)"\s\(K')
    _hotweek_pattern = re.compile(r"Extreme Hot Week Period selected:"
                                  r"\s*(\w{3})\s*(\d{1,2}):\s*(\w{3})\s*(\d{1,2}),")
    _coldweek_pattern = re.compile(r"Extreme Cold Week Period selected:"
                                   r"\s*(\w{3})\s*(\d{1,2}):\s*(\w{3})\s*(\d{1,2}),")
    _typweek_pattern = re.compile(r"(\S*)\s*Typical Week Period selected:"
                                  r"\s*(\w{3})\s*(\d{1,2}):\s*(\w{3})\s*(\d{1,2}),")
    _heat_pattern = re.compile(r"Heating\s(\d.*)")
    _cool_pattern = re.compile(r"Cooling\s(\d.*)")
    _tau_beam_pattern = re.compile(r"taub \(beam\)(.*)")
    _tau_diffuse_pattern = re.compile(r"taud \(diffuse\)(.*)")
    _db_50_pattern = re.compile(r"Drybulb 5.0%(.*)")
    _wb_50_pattern = re.compile(r"Coincident Wetbulb 5.0%(.*)")
    _db_100_pattern = re.compile(r"Drybulb 10.%(.*)")
    _wb_100_pattern = re.compile(r"Coincident Wetbulb 10.%(.*)")
    _db_20_pattern = re.compile(r"Drybulb 2.0%(.*)")
    _wb_20_pattern = re.compile(r"Coincident Wetbulb 2.0%(.*)")
    _db_04_pattern = re.compile(r"Drybulb 0.4%(.*)")
    _wb_04_pattern = re.compile(r"Coincident Wetbulb 0.4%(.*)")
    _db_range_50_pattern = re.compile(r"Drybulb range - DB 5%(.*)")
    _wb_range_50_pattern = re.compile(r"Wetbulb range - DB 5%(.*)")
    _winds_pattern = re.compile(r"Monthly Statistics for Wind Speed[\s\S]*Daily Avg(.*)")
    _windd_patterns = tuple(re.compile(
        r"Monthly Wind Direction %[\s\S]*" + dir + r"\s(.*)") for dir in _wind_dir_names)

    __slots__ = ('_file_path', '_winter_des_day_dict', '_summer_des_day_dict',
                 '_monthly_wind_dirs', '_location', '_ashrae_climate_zone',
                 '_koppen_climate_zone', '_extreme_cold_week', '_extreme_hot_week',
                 '_typical_weeks', '_monthly_db_50', '_monthly_wb_50', '_monthly_db_range_50',
                 '_monthly_wb_range_50', '_monthly_db_100', '_monthly_wb_100', '_monthly_db_20',
                 '_monthly_wb_20', '_monthly_db_04', '_monthly_wb_04', '_monthly_wind',
                 '_stand_press_at_elev', '_monthly_tau_beam', '_monthly_tau_diffuse',
                 '_header', '_body')

    def __init__(self, file_path):
        """Initialize the class.
        """
        if file_path is not None:
            if not os.path.isfile(file_path):
                raise ValueError(
                    'Cannot find an stat file at {}'.format(file_path))
            if not file_path.lower().endswith('stat'):
                raise TypeError('{} is not an .stat file.'.format(file_path))
            self._file_path = os.path.normpath(file_path)

        # defaults empty state for certain parameters
        self._winter_des_day_dict = {}
        self._summer_des_day_dict = {}
        self._monthly_wind_dirs = []

        # import the data from the file
        if file_path is not None:
            self._import_data()

    @classmethod
    def from_dict(cls, data):
        """ Create Stat from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

                {
                "location": {},  # ladybug location schema
                "ashrae_climate_zone": ""5A,  # str
                "koppen_climate_zone": "Dfa", # str
                "extreme_cold_week": {},  # ladybug analysis period schema
                "extreme_hot_week": {},  # ladybug analysis period schema
                "typical_weeks": {},  # dict of ladybug analysis period schemas
                "heating_dict": {},  # dict containing heating design conditions
                "cooling_dict": {},  # dict containing cooling design conditions
                "monthly_db_50": [],  # list of 12 float values for each month
                "monthly_wb_50": [],  # list of 12 float values for each month
                "monthly_db_range_50": [],  # list of 12 float values for each month
                "monthly_wb_range_50": [],  # list of 12 float values for each month
                "monthly_db_100": [],  # list of 12 float values for each month
                "monthly_wb_100": [],  # list of 12 float values for each month
                "monthly_db_20": [],  # list of 12 float values for each month
                "monthly_wb_20": [],  # list of 12 float values for each month
                "monthly_db_04": [],  # list of 12 float values for each month
                "monthly_wb_04": [],  # list of 12 float values for each month
                "monthly_wind": [],  # list of 12 float values for each month
                "monthly_wind_dirs": [],  # matrix with 12 cols for months of the year
                                          #and 8 rows for the cardinal directions.
                "standard_pressure_at_elev": 101325,  # float value for pressure in Pa
                "monthly_tau_beam":[],  # list of 12 float values for each month
                "monthly_tau_diffuse": []  # list of 12 float values for each month
                }
        """
        # Initialize the class with all data missing
        stat_ob = cls(None)

        # Check required and optional keys
        option_keys_none = ('ashrae_climate_zone', 'koppen_climate_zone',
                            'extreme_cold_week', 'extreme_hot_week',
                            'standard_pressure_at_elev')
        option_keys_list = ('monthly_db_50', 'monthly_wb_50',
                            'monthly_db_range_50', 'monthly_wb_range_50',
                            'monthly_db_100', 'monthly_wb_100', 'monthly_db_20',
                            'monthly_wb_20', 'monthly_db_04', 'monthly_wb_04',
                            'monthly_wind', 'monthly_wind_dirs',
                            'monthly_tau_beam', 'monthly_tau_diffuse')
        option_keys_dict = ('typical_weeks', 'heating_dict', 'cooling_dict')
        assert 'location' in data, 'Required key "location" is missing!'
        for key in option_keys_none:
            if key not in data:
                data[key] = None
        for key in option_keys_list:
            if key not in data:
                data[key] = []
        for key in option_keys_dict:
            if key not in data:
                data[key] = {}

        # assign the properties of the dictionary to the stat object.
        stat_ob._location = Location.from_dict(data['location'])
        stat_ob._ashrae_climate_zone = data['ashrae_climate_zone']
        stat_ob._koppen_climate_zone = data['koppen_climate_zone']
        stat_ob._extreme_cold_week = AnalysisPeriod.from_dict(data['extreme_cold_week'])\
            if data['extreme_cold_week'] else None
        stat_ob._extreme_hot_week = AnalysisPeriod.from_dict(data['extreme_hot_week'])\
            if data['extreme_hot_week'] else None
        stat_ob._typical_weeks = {}
        for key, val in data['typical_weeks'].items():
            if isinstance(val, list):
                stat_ob._typical_weeks[key] = [AnalysisPeriod.from_dict(v) for v in val]
            else:
                stat_ob._typical_weeks[key] = AnalysisPeriod.from_dict(val)
        stat_ob._winter_des_day_dict = data['heating_dict']
        stat_ob._summer_des_day_dict = data['cooling_dict']
        stat_ob._monthly_db_50 = data['monthly_db_50']
        stat_ob._monthly_wb_50 = data['monthly_wb_50']
        stat_ob._monthly_db_range_50 = data['monthly_db_range_50']
        stat_ob._monthly_wb_range_50 = data['monthly_wb_range_50']
        stat_ob._monthly_db_100 = data['monthly_db_100']
        stat_ob._monthly_wb_100 = data['monthly_wb_100']
        stat_ob._monthly_db_20 = data['monthly_db_20']
        stat_ob._monthly_wb_20 = data['monthly_wb_20']
        stat_ob._monthly_db_04 = data['monthly_db_04']
        stat_ob._monthly_wb_04 = data['monthly_wb_04']
        stat_ob._monthly_wind = data['monthly_wind']
        stat_ob._monthly_wind_dirs = data['monthly_wind_dirs']
        stat_ob._stand_press_at_elev = data['standard_pressure_at_elev']
        stat_ob._monthly_tau_beam = data['monthly_tau_beam']
        stat_ob._monthly_tau_diffuse = data['monthly_tau_diffuse']

        return stat_ob

    @property
    def file_path(self):
        """Get the path to the stat file."""
        return self._file_path

    def _import_data(self):
        """Import data from a stat file.
        """
        # set default state to ironpython for very old ironpython (2.7.0)
        iron_python = True
        try:
            iron_python = True if platform.python_implementation() == 'IronPython' \
                else False
        except ValueError as e:
            # older versions of IronPython fail to parse version correctly
            # failed to parse IronPython sys.version: '2.7.5 (IronPython 2.7.5 (2.7.5.0)
            # on .NET 4.0.30319.42000 (64-bit))'
            if 'IronPython' in str(e):
                iron_python = True

        if iron_python:
            statwin = codecs.open(self.file_path, 'r')
        else:
            statwin = codecs.open(self.file_path, 'r', encoding='utf-8', errors='ignore')
        try:
            line = statwin.readline()
            # import header with location
            self._header = [line] + [statwin.readline() for i in xrange(9)]
            self._body = statwin.read()
        except Exception as e:
            import traceback
            raise Exception('{}\n{}'.format(e, traceback.format_exc()))
        else:

            # import location data
            loc_name = self._header[2].strip().replace('Location -- ', '')
            if ' - ' in loc_name:
                city = ' '.join(loc_name.split(' - ')[:-1])
            else:  # for US stat files it is full name separated by spaces
                city = ' '.join(loc_name.split()[:-2])
            country = loc_name.split(' ')[-1]
            source = self._header[6].strip().replace('Data Source -- ', '')
            station_id = self._header[8].strip().replace('WMO Station ', '')
            if iron_python:  # IronPython
                matches = self._coord_pattern1.findall(
                    self._header[3].replace('\xb0', 'deg'))
            else:  # CPython
                matches = self._coord_pattern2.findall(self._header[3])
            lat_sign = -1 if matches[0][0] == 'S' else 1
            latitude = lat_sign * (float(matches[0][1]) + (float(matches[0][2]) / 60))
            lon_sign = -1 if matches[1][0] == 'W' else 1
            longitude = lon_sign * (float(matches[1][1]) + (float(matches[1][2]) / 60))
            time_zone = self._regex_check(self._timez_pattern, self._header[3])
            elev_matches = self._elev_pattern1.findall(self._header[4])
            if len(elev_matches) == 0:
                elev_matches = self._elev_pattern2.findall(self._header[4])
            elev_sign = -1 if elev_matches[0][-1].lower() == 'below' else 1
            elevation = elev_sign * float(elev_matches[0][0])
            self._location = Location()
            self._location.city = city
            self._location.country = country
            self._location.source = source
            self._location.station_id = station_id
            self._location.latitude = latitude
            self._location.longitude = longitude
            self._location.time_zone = time_zone
            self._location.elevation = elevation

            # pull out individual properties
            self._stand_press_at_elev = self._regex_check(
                self._press_pattern, self._header[5])
            self._ashrae_climate_zone = self._regex_check(
                self._ashraecz_pattern, self._body)
            self._koppen_climate_zone = self._regex_check(
                self._koppencz_pattern, self._body)

            # pull out extreme and seasonal weeks.
            self._extreme_hot_week = self._regex_week_parse(self._hotweek_pattern)
            self._extreme_cold_week = self._regex_week_parse(self._coldweek_pattern)
            self._typical_weeks = self._regex_typical_week_parse()

            # pull out annual design days
            winter_vals = self._regex_parse(self._heat_pattern)
            for key, val in zip(DesignDay.HEATING_KEYS, winter_vals):
                self._winter_des_day_dict[key] = val
            summer_vals = self._regex_parse(self._cool_pattern)
            for key, val in zip(DesignDay.COOLING_KEYS, summer_vals):
                self._summer_des_day_dict[key] = val

            # Pull out relevant monthly information
            self._monthly_tau_beam = self._regex_parse(self._tau_beam_pattern)
            self._monthly_tau_diffuse = self._regex_parse(self._tau_diffuse_pattern)
            self._monthly_db_50 = self._regex_parse(self._db_50_pattern)
            self._monthly_wb_50 = self._regex_parse(self._wb_50_pattern)
            self._monthly_db_100 = self._regex_parse(self._db_100_pattern)
            self._monthly_wb_100 = self._regex_parse(self._wb_100_pattern)
            self._monthly_db_20 = self._regex_parse(self._db_20_pattern)
            self._monthly_wb_20 = self._regex_parse(self._wb_20_pattern)
            self._monthly_db_04 = self._regex_parse(self._db_04_pattern)
            self._monthly_wb_04 = self._regex_parse(self._wb_04_pattern)
            self._monthly_db_range_50 = self._regex_parse(self._db_range_50_pattern)
            self._monthly_wb_range_50 = self._regex_parse(self._wb_range_50_pattern)
            self._monthly_wind = self._regex_parse(self._winds_pattern)
            for direction in self._windd_patterns:
                dirs = self._regex_parse(direction)
                if dirs != []:
                    self._monthly_wind_dirs.append(dirs)
            if self._monthly_wind_dirs == []:
                self._monthly_wind_dirs = [[0] * 12 for i in xrange(8)]

        finally:
            statwin.close()

    def _regex_check(self, regex_pattern, search_space):
        matches = regex_pattern.findall(search_space)
        if len(matches) > 0:
            try:
                return float(matches[0])
            except ValueError:
                return matches[0]
        else:
            return None

    def _regex_week(self, match):
        if len(match) == 4:
            try:
                st_mon = int(self._months.index(match[0])) + 1
                end_mon = int(self._months.index(match[2])) + 1
                st_day = int(match[1])
                end_day = int(match[3])
            except ValueError:
                return None
            return AnalysisPeriod(st_mon, st_day, 0, end_mon, end_day, 23)
        else:
            return None

    def _regex_week_parse(self, regex_pattern):
        matches = regex_pattern.findall(self._body)
        if len(matches) > 0:
            return self._regex_week(matches[0])
        else:
            return None

    def _regex_typical_week_parse(self):
        typ_weeks = {'other': []}
        matches = self._typweek_pattern.findall(self._body)
        for match in matches:
            a_per = self._regex_week(match[1:])
            if 'winter' in match[0]:
                typ_weeks['winter'] = a_per
            elif 'spring' in match[0]:
                typ_weeks['spring'] = a_per
            elif 'summer' in match[0]:
                typ_weeks['summer'] = a_per
            elif 'autumn' in match[0]:
                typ_weeks['autumn'] = a_per
            else:
                typ_weeks['other'].append(a_per)
        return typ_weeks

    def _regex_parse(self, regex_pattern):
        matches = regex_pattern.findall(self._body)
        if len(matches) > 0:
            raw_txt = matches[0].strip().split('\t')
            try:
                return [float(i) if i != 'N' else None for i in raw_txt]
            except ValueError:
                return [str(i) for i in raw_txt]
        else:
            return []

    @property
    def monthly_found(self):
        if self._monthly_db_range_50 != [] and self._monthly_wb_range_50 != [] \
            and self._monthly_wind != [] \
                and self._stand_press_at_elev is not None:
            return True
        else:
            return False

    @property
    def location(self):
        """Return ladybug location object."""
        return self._location

    @property
    def ashrae_climate_zone(self):
        """Return a text string indicating the ASHRAE climate zone.

        Numbers in the zone denote average temperature (0 = Hottest; 8 = Coldest)
        Letters in the zone denote wetness (A = Humid; B = Dry; C = Marine)
        """
        return self._ashrae_climate_zone

    @property
    def koppen_climate_zone(self):
        """Return a text string indicating the Koppen climate zone.

        The Koppen climate classification is the most widely used climate
        classification system and combines average annual and monthly
        temperatures, precipitation, and the seasonality of precipitation.
        """
        return self._koppen_climate_zone

    @property
    def extreme_cold_week(self):
        """AnalysisPeriod for the coldest week within the corresponding EPW."""
        return self._extreme_cold_week

    @property
    def extreme_hot_week(self):
        """AnalysisPeriod for the hottest week within the corresponding EPW."""
        return self._extreme_hot_week

    @property
    def typical_winter_week(self):
        """AnalysisPeriod for a typical winter week within the corresponding EPW."""
        try:
            return self._typical_weeks['winter']
        except KeyError:
            return None

    @property
    def typical_spring_week(self):
        """AnalysisPeriod for a typical spring week within the corresponding EPW."""
        try:
            return self._typical_weeks['spring']
        except KeyError:
            return None

    @property
    def typical_summer_week(self):
        """AnalysisPeriod for a typical summer week within the corresponding EPW."""
        try:
            return self._typical_weeks['summer']
        except KeyError:
            return None

    @property
    def typical_autumn_week(self):
        """AnalysisPeriod for a typical autumn week within the corresponding EPW."""
        try:
            return self._typical_weeks['autumn']
        except KeyError:
            return None

    @property
    def other_typical_weeks(self):
        """List of AnalysisPeriods for typical weeks outside of the seasonal weeks."""
        return self._typical_weeks['other']

    @property
    def annual_heating_design_day_996(self):
        """A design day object representing the annual 99.6% heating design day."""
        if bool(self._winter_des_day_dict):
            return DesignDay.from_ashrae_dict_heating(
                self._winter_des_day_dict, self.location, False,
                self._stand_press_at_elev)
        else:
            return None

    @property
    def annual_heating_design_day_990(self):
        """A design day object representing the annual 99.0% heating design day."""
        if bool(self._winter_des_day_dict):
            return DesignDay.from_ashrae_dict_heating(
                self._winter_des_day_dict, self.location, True,
                self._stand_press_at_elev)
        else:
            return None

    @property
    def annual_cooling_design_day_004(self):
        """A design day object representing the annual 0.4% cooling design day."""
        if bool(self._summer_des_day_dict):
            tau = None
            month_num = int(self._summer_des_day_dict['Month'])
            if self._monthly_tau_beam != [] and self._monthly_tau_diffuse != [] \
                and self._monthly_tau_beam[month_num - 1] is not None and \
                    self._monthly_tau_diffuse[month_num - 1] is not None:
                tau = (self._monthly_tau_beam[month_num - 1],
                       self._monthly_tau_diffuse[month_num - 1])
            return DesignDay.from_ashrae_dict_cooling(
                self._summer_des_day_dict, self.location, False,
                self._stand_press_at_elev, tau)
        else:
            return None

    @property
    def annual_cooling_design_day_010(self):
        """A design day object representing the annual 1.0% cooling design day."""
        if bool(self._summer_des_day_dict):
            tau = None
            month_num = int(self._summer_des_day_dict['Month'])
            if self._monthly_tau_beam != [] and self._monthly_tau_diffuse != [] \
                and self._monthly_tau_beam[month_num - 1] is not None and \
                    self._monthly_tau_diffuse[month_num - 1] is not None:
                tau = (self._monthly_tau_beam[month_num - 1],
                       self._monthly_tau_diffuse[month_num - 1])
            return DesignDay.from_ashrae_dict_cooling(
                self._summer_des_day_dict, self.location, True,
                self._stand_press_at_elev, tau)
        else:
            return None

    @property
    def monthly_cooling_design_days_050(self):
        """A list of 12 objects representing monthly 5.0% cooling design days."""
        if not self.monthly_found or self._monthly_db_50 == [] \
                or self._monthly_wb_50 == []:
            return []
        else:
            db_conds = [DryBulbCondition(x, y) for x, y in zip(
                self._monthly_db_50, self._monthly_db_range_50)]
            hu_conds = [HumidityCondition(
                'Wetbulb', x, self._stand_press_at_elev) for x in self._monthly_wb_50]
            ws_conds = self.monthly_wind_conditions
            sky_conds = self.monthly_clear_sky_conditions
            return [DesignDay(
                '5% Cooling Design Day for {}'.format(self._months[i]),
                'SummerDesignDay', self._location,
                db_conds[i], hu_conds[i], ws_conds[i], sky_conds[i])
                    for i in xrange(12)]

    @property
    def monthly_cooling_design_days_100(self):
        """A list of 12 objects representing monthly 10.0% cooling design days."""
        if not self.monthly_found or self._monthly_db_100 == [] \
                or self._monthly_wb_100 == []:
            return []
        else:
            db_conds = [DryBulbCondition(x, y) for x, y in zip(
                self._monthly_db_100, self._monthly_db_range_50)]
            hu_conds = [HumidityCondition(
                'Wetbulb', x, self._stand_press_at_elev) for x in self._monthly_wb_100]
            ws_conds = self.monthly_wind_conditions
            sky_conds = self.monthly_clear_sky_conditions
            return [DesignDay(
                '10% Cooling Design Day for {}'.format(self._months[i]),
                'SummerDesignDay', self._location,
                db_conds[i], hu_conds[i], ws_conds[i], sky_conds[i])
                    for i in xrange(12)]

    @property
    def monthly_cooling_design_days_020(self):
        """A list of 12 objects representing monthly 2.0% cooling design days."""
        if not self.monthly_found or self._monthly_db_20 == [] \
                or self._monthly_wb_20 == []:
            return []
        else:
            db_conds = [DryBulbCondition(x, y) for x, y in zip(
                self._monthly_db_20, self._monthly_db_range_50)]
            hu_conds = [HumidityCondition(
                'Wetbulb', x, self._stand_press_at_elev) for x in self._monthly_wb_20]
            ws_conds = self.monthly_wind_conditions
            sky_conds = self.monthly_clear_sky_conditions
            return [DesignDay(
                '2% Cooling Design Day for {}'.format(self._months[i]),
                'SummerDesignDay', self._location,
                db_conds[i], hu_conds[i], ws_conds[i], sky_conds[i])
                    for i in xrange(12)]

    @property
    def monthly_cooling_design_days_004(self):
        """A list of 12 objects representing monthly 0.4% cooling design days."""
        if not self.monthly_found or self._monthly_db_04 == [] \
                or self._monthly_wb_04 == []:
            return []
        else:
            db_conds = [DryBulbCondition(x, y) for x, y in zip(
                self._monthly_db_04, self._monthly_db_range_50)]
            hu_conds = [HumidityCondition(
                'Wetbulb', x, self._stand_press_at_elev) for x in self._monthly_wb_04]
            ws_conds = self.monthly_wind_conditions
            sky_conds = self.monthly_clear_sky_conditions
            return [DesignDay(
                '0.4% Cooling Design Day for {}'.format(self._months[i]),
                'SummerDesignDay', self._location,
                db_conds[i], hu_conds[i], ws_conds[i], sky_conds[i])
                    for i in xrange(12)]

    @property
    def monthly_db_temp_050(self):
        """A list of 12 float values for monthly 5.0% dry bulb temperature."""
        return self._monthly_db_50

    @property
    def monthly_wb_temp_050(self):
        """A list of 12 float values for monthly 5.0% wet bulb temperature."""
        return self._monthly_wb_50

    @property
    def monthly_db_temp_range_050(self):
        """A list of 12 values for monthly ranges of dry bulb temperatures at 5.0%."""
        return self._monthly_db_range_50

    @property
    def monthly_wb_temp_range_050(self):
        """A list of 12 values for monthly ranges of wet bulb temperatures at 5.0%."""
        return self._monthly_wb_range_50

    @property
    def standard_pressure_at_elev(self):
        """The standard pressure on pascals at the elevation of the location."""
        return self._stand_press_at_elev

    @property
    def monthly_wind_conditions(self):
        """A list of 12 monthly wind conditions that are used on the design days."""
        return [WindCondition(x, y) for x, y in zip(
            self._monthly_wind, self.monthly_wind_dirs)]

    @property
    def monthly_ws_avg(self):
        """A list of 12 float values for monthly average wind speeds."""
        return self._monthly_wind

    @property
    def monthly_wind_dirs(self):
        """A list of prevailing wind directions for each month."""
        mwd = zip(*self._monthly_wind_dirs)
        return [self._wind_dirs[mon.index(max(mon))] for mon in mwd]

    @property
    def monthly_clear_sky_conditions(self):
        """A list of 12 monthly clear sky conditions that are used on the design days."""
        if self._monthly_tau_diffuse is [] or self._monthly_tau_beam is []:
            return [ASHRAEClearSky(Date(i, 21)) for i in xrange(1, 13)]
        return [ASHRAETau(Date(i, 21), x, y) for i, x, y in zip(
            list(xrange(1, 13)), self._monthly_tau_beam, self._monthly_tau_diffuse)]

    @property
    def monthly_tau_beam(self):
        """A list of 12 float values for monthly beam optical depth.

        These values can be used to generate ASHRAE Revised Clear Skies, which
        are intended to determine peak solar load and sizing parmeters for
        HVAC systems.
        """
        return self._monthly_tau_beam

    @property
    def monthly_tau_diffuse(self):
        """Return a list of 12 float values for monthly diffuse optical depth.

        These values can be used to generate ASHRAE Revised Clear Skies, which
        are intended to determine peak solar load and sizing parmeters for
        HVAC systems.
        """
        return self._monthly_tau_diffuse

    def to_dict(self):
        """Convert the stat object to a dictionary."""
        def dictify_dict(base_dict):
            new_dict = {}
            for key, val in base_dict.items():
                if isinstance(val, list):
                    new_dict[key] = [v.to_dict() for v in val]
                else:
                    new_dict[key] = val.to_dict()
            return new_dict
        return {
            'location': self.location.to_dict(),
            'ashrae_climate_zone': self.ashrae_climate_zone,
            'koppen_climate_zone': self.koppen_climate_zone,
            'extreme_cold_week': self.extreme_cold_week.to_dict()
            if self.extreme_cold_week else None,
            'extreme_hot_week': self.extreme_hot_week.to_dict()
            if self.extreme_cold_week else None,
            'typical_weeks': dictify_dict(self._typical_weeks),
            'heating_dict': self._winter_des_day_dict,
            'cooling_dict': self._summer_des_day_dict,
            "monthly_db_50": self._monthly_db_50,
            "monthly_wb_50": self._monthly_wb_50,
            "monthly_db_range_50": self._monthly_db_range_50,
            "monthly_wb_range_50": self._monthly_wb_range_50,
            "monthly_db_100": self._monthly_db_100,
            "monthly_wb_100": self._monthly_wb_100,
            "monthly_db_20": self._monthly_db_20,
            "monthly_wb_20": self._monthly_wb_20,
            "monthly_db_04": self._monthly_db_04,
            "monthly_wb_04": self._monthly_wb_04,
            "monthly_wind": self._monthly_wind,
            "monthly_wind_dirs": self._monthly_wind_dirs,
            "standard_pressure_at_elev": self.standard_pressure_at_elev,
            "monthly_tau_beam": self.monthly_tau_beam,
            "monthly_tau_diffuse": self.monthly_tau_diffuse,
            "type": 'STAT'
        }

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """stat file representation."""
        return "STAT [%s]" % self.location.city



# coding=utf-8
from ladybug.dt import DateTime, Date, Time
import pickle


def test_date_time_init():
    """Test the init method for DateTime and basic properties."""
    dt1 = DateTime(6, 21, 12)
    dt2 = DateTime(6, 21, 12)
    dt3 = DateTime(6, 21, 12, leap_year=True)
    dt4 = DateTime(6, 21, 13)
    str(dt1)  # test the string representation of Datetime

    assert dt1.month == 6
    assert dt1.day == 21
    assert dt1.hour == 12
    assert dt1.minute == 0
    assert not dt1.leap_year
    assert dt1.doy == 172
    assert dt1.hoy == 4116
    assert dt1.moy == 246960
    assert isinstance(dt1.int_hoy, int)
    assert dt1.int_hoy == 4116
    assert dt1.float_hour == 12.0
    assert dt1.date == Date(6, 21)
    assert dt1.time == Time(12, 0)

    assert dt1 == dt2
    assert dt1 != dt3
    assert dt1 != dt4

    assert sorted([dt4, dt1]) == [dt1, dt4]


def test_date_time_to_from_dict():
    """Test the dict methods for DateTime."""
    dt1 = DateTime(6, 21, 12)
    dt_dict = dt1.to_dict()
    rebuilt_dt = DateTime.from_dict(dt_dict)
    assert dt1 == rebuilt_dt
    assert rebuilt_dt.to_dict() == dt_dict


def test_to_from_date_time_string():
    """Test the from_date_time_string method for DateTime."""
    dt1 = DateTime(6, 21, 12)
    dt_str = str(dt1)
    rebuilt_dt = DateTime.from_date_time_string(dt_str)
    assert rebuilt_dt == dt1
    assert str(rebuilt_dt) == dt_str


def test_to_from_array():
    """Test the from_array method for DateTime."""
    dt1 = DateTime(6, 21, 12)
    dt_arr = dt1.to_array()
    rebuilt_dt = DateTime.from_array(dt_arr)
    assert rebuilt_dt == dt1
    assert rebuilt_dt.to_array() == dt_arr


def test_date_time_from_hoy():
    """Test the from_hoy method for DateTime and basic properties."""
    dt1 = DateTime.from_hoy(4116)
    assert dt1 == DateTime(6, 21, 12)
    dt2 = DateTime.from_hoy(4116, leap_year=True)
    assert dt2 == DateTime(6, 20, 12, leap_year=True)


def test_date_time_from_moy():
    """Test the from_moy method for DateTime and basic properties."""
    dt1 = DateTime.from_moy(246960)
    assert dt1 == DateTime(6, 21, 12)
    dt2 = DateTime.from_moy(246960, leap_year=True)
    assert dt2 == DateTime(6, 20, 12, leap_year=True)


def test_date_time_add_sub():
    """Test the add and subtract methods for DateTime."""
    dt1 = DateTime(6, 21, 12)
    dt2 = dt1.add_hour(1)
    dt3 = dt1.sub_hour(1)
    dt4 = dt1.add_minute(1)
    dt5 = dt1.sub_minute(1)

    assert dt2 == DateTime(6, 21, 13)
    assert dt3 == DateTime(6, 21, 11)
    assert dt4 == DateTime(6, 21, 12, 1)
    assert dt5 == DateTime(6, 21, 11, 59)


def test_date_init():
    """Test the init method for Date and basic properties."""
    dt1 = Date(6, 21)
    dt2 = Date(6, 21)
    dt3 = Date(6, 21, leap_year=True)
    dt4 = Date(6, 22)
    str(dt1)  # test the string representation of Date

    assert dt1.month == 6
    assert dt1.day == 21
    assert not dt1.leap_year
    assert dt1.doy == 172

    assert dt1 == dt2
    assert dt1 != dt3
    assert dt1 != dt4

    assert sorted([dt4, dt1]) == [dt1, dt4]


def test_date_from_doy():
    """Test the from_doy method for Date and basic properties."""
    dt1 = Date.from_doy(172)
    assert dt1 == Date(6, 21)
    dt2 = Date.from_doy(172, leap_year=True)
    assert dt2 == Date(6, 20, leap_year=True)
    dt3 = Date.from_doy(181)
    assert dt3 == Date(6, 30)
    dt4 = Date.from_doy(182)
    assert dt4 == Date(7, 1)


def test_date_to_from_dict():
    """Test the dict methods for Date."""
    dt1 = Date(6, 21)
    dt_dict = dt1.to_dict()
    rebuilt_dt = Date.from_dict(dt_dict)
    assert dt1 == rebuilt_dt
    assert rebuilt_dt.to_dict() == dt_dict


def test_to_from_date_string():
    """Test the from_date_string method for Date."""
    dt1 = Date(6, 21)
    dt_str = str(dt1)
    rebuilt_dt = Date.from_date_string(dt_str)
    assert rebuilt_dt == dt1
    assert str(rebuilt_dt) == dt_str


def test_date_to_from_array():
    """Test the from_array method for Date."""
    dt1 = Date(6, 21)
    dt_arr = dt1.to_array()
    rebuilt_dt = Date.from_array(dt_arr)
    assert rebuilt_dt == dt1
    assert rebuilt_dt.to_array() == dt_arr


def test_time_init():
    """Test the init method for Date and basic properties."""
    t1 = Time(12, 30)
    t2 = Time(12, 30)
    t3 = Time(12, 31)
    str(t1)  # test the string representation of Time

    assert t1.hour == 12
    assert t1.minute == 30
    assert t1.mod == 750

    assert t1 == t2
    assert t1 != t3

    assert sorted([t3, t1]) == [t1, t3]


def test_time_from_mod():
    """Test the from_mod method for Time and basic properties."""
    t1 = Time.from_mod(750)
    assert t1 == Time(12, 30)


def test_time_to_from_dict():
    """Test the dict methods for Time."""
    t1 = Time(12, 30)
    t_dict = t1.to_dict()
    rebuilt_t = Time.from_dict(t_dict)
    assert t1 == rebuilt_t
    assert rebuilt_t.to_dict() == t_dict


def test_to_from_time_string():
    """Test the from_time_string method for Time."""
    t1 = Time(12, 30)
    t_str = str(t1)
    rebuilt_t = Time.from_time_string(t_str)
    assert rebuilt_t == t1
    assert str(rebuilt_t) == t_str


def test_time_to_from_array():
    """Test the from_array method for Date."""
    t1 = Time(12, 30)
    t_arr = t1.to_array()
    rebuilt_t = Time.from_array(t_arr)
    assert rebuilt_t == t1
    assert rebuilt_t.to_array() == t_arr


def test_pickle_and_unpickle():
    """Test the pickling and unpickling of DateTime, Date, and Time objects."""
    dt1 = DateTime(6, 21, 12)
    dt2 = Date(6, 21)
    dt3 = Time(12, 30)

    serialized_dt1 = pickle.dumps(dt1)
    serialized_dt2 = pickle.dumps(dt2)
    serialized_dt3 = pickle.dumps(dt3)

    assert pickle.loads(serialized_dt1) == dt1
    assert pickle.loads(serialized_dt2) == dt2
    assert pickle.loads(serialized_dt3) == dt3



# coding=utf-8
from __future__ import division

from ladybug.legend import Legend, LegendParameters, LegendParametersCategorized
from ladybug.color import Color, Colorset, ColorRange
from ladybug.datatype.thermalcondition import PredictedMeanVote

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry2d.mesh import Mesh2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.mesh import Mesh3D

import pytest


def test_init_legend_parameters():
    """Test the initialization of LegendParameter objects."""
    leg_par = LegendParameters(0, 1000)
    str(leg_par)  # Test the LegendParameters representation

    assert leg_par.min == 0
    assert leg_par.max == 1000
    assert leg_par.is_segment_count_default
    assert leg_par.is_title_default
    assert leg_par.is_base_plane_default
    assert leg_par.is_segment_height_default
    assert leg_par.is_segment_width_default
    assert leg_par.is_text_height_default

    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.min == leg_par.min
    assert leg_par_copy.max == leg_par.max
    assert leg_par_copy.segment_count == leg_par.segment_count


def test_to_from_dict():
    """Test the to/from dict methods."""
    leg_par = LegendParameters(0, 1000)
    leg_par_dict = leg_par.to_dict()
    new_leg_par = LegendParameters.from_dict(leg_par_dict)
    assert new_leg_par.to_dict() == leg_par_dict


def test_colors():
    """Test the LegendParameter colors property."""
    leg_par = LegendParameters(colors=[Color(0, 0, 0), Color(255, 255, 255)])

    assert len(leg_par.colors) == 2
    assert leg_par.colors[0] == Color(0, 0, 0)
    assert leg_par.colors[1] == Color(255, 255, 255)

    leg_par_copy = leg_par.duplicate()
    assert len(leg_par_copy.colors) == 2
    assert leg_par_copy.colors[0] == Color(0, 0, 0)
    assert leg_par_copy.colors[1] == Color(255, 255, 255)

    leg_par.colors = [Color(0, 0, 0), Color(100, 100, 100)]
    assert leg_par.colors[1] == Color(100, 100, 100)

    with pytest.raises(Exception):
        leg_par = LegendParameters(colors=[0, 1])
    with pytest.raises(Exception):
        leg_par.colors = [0, 1]


def test_continuous_legend():
    """Test the LegendParameter continuous_legend property."""
    leg_par = LegendParameters()
    leg_par.continuous_legend = True

    assert leg_par.continuous_legend
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.continuous_legend

    leg_par.continuous_legend = False
    assert not leg_par.continuous_legend

    with pytest.raises(Exception):
        leg_par = LegendParameters(continuous_legend='yes')
    with pytest.raises(Exception):
        leg_par.continuous_legend = 'yes'


def test_title():
    """Test the LegendParameter title property."""
    leg_par = LegendParameters(title='m2')

    assert leg_par.title == 'm2'
    assert not leg_par.is_title_default
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.title == 'm2'

    leg_par.title = 'm2'
    assert leg_par.title == 'm2'

    with pytest.raises(Exception):
        leg_par = LegendParameters(title=2)
    with pytest.raises(Exception):
        leg_par.title = 2


def test_ordinal_dictionary():
    """Test the LegendParameter ordinal_dictionary property."""
    leg_par = LegendParameters(min=0, max=1)
    leg_par.ordinal_dictionary = {0: 'False', 1: 'True'}

    assert leg_par.ordinal_dictionary == {0: 'False', 1: 'True'}
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.ordinal_dictionary == {0: 'False', 1: 'True'}

    leg_par.ordinal_dictionary = {0: 'No', 1: 'Yes'}
    assert leg_par.ordinal_dictionary == {0: 'No', 1: 'Yes'}

    with pytest.raises(Exception):
        leg_par = LegendParameters(ordinal_dictionary=['False', 'True'])
    with pytest.raises(Exception):
        leg_par.ordinal_dictionary = ['False', 'True']


def test_decimal_count():
    """Test the LegendParameter decimal_count property."""
    leg_par = LegendParameters()
    leg_par.decimal_count = 3

    assert leg_par.decimal_count == 3
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.decimal_count == 3

    leg_par.decimal_count = 0
    assert leg_par.decimal_count == 0

    with pytest.raises(Exception):
        leg_par = LegendParameters(decimal_count='2')
    with pytest.raises(Exception):
        leg_par.decimal_count = '2'


def test_vertical():
    """Test the LegendParameter continuous_legend property."""
    leg_par = LegendParameters()
    leg_par.vertical = False

    assert not leg_par.vertical
    leg_par_copy = leg_par.duplicate()
    assert not leg_par_copy.vertical

    leg_par.vertical = True
    assert leg_par.vertical

    with pytest.raises(Exception):
        leg_par = LegendParameters(vertical='yes')
    with pytest.raises(Exception):
        leg_par.vertical = 'yes'


def test_base_plane():
    """Test the LegendParameter base_plane property."""
    leg_par = LegendParameters(
        base_plane=Plane(Vector3D(0, 0, 0), Point3D(10, 10, 0)))

    assert leg_par.base_plane.o == Point3D(10, 10, 0)
    assert not leg_par.is_base_plane_default
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.base_plane.o == Point3D(10, 10, 0)

    leg_par.base_plane = Plane(Vector3D(0, 0, 0), Point3D(-10, -10, 0))
    assert leg_par.base_plane.o == Point3D(-10, -10, 0)

    with pytest.raises(Exception):
        leg_par = LegendParameters(base_plane=Point3D(-10, -10, 0))
    with pytest.raises(Exception):
        leg_par.base_plane = Point3D(-10, -10, 0)


def test_segment_height():
    """Test the LegendParameter segment_height property."""
    leg_par = LegendParameters()
    leg_par.segment_height = 3

    assert leg_par.segment_height == 3
    assert not leg_par.is_segment_height_default
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.segment_height == 3

    leg_par.segment_height = 2
    assert leg_par.segment_height == 2

    with pytest.raises(Exception):
        leg_par = LegendParameters(segment_height=0)
    with pytest.raises(Exception):
        leg_par.segment_height = 0


def test_segment_width():
    """Test the LegendParameter segment_width property."""
    leg_par = LegendParameters()
    leg_par.segment_width = 3

    assert leg_par.segment_width == 3
    assert not leg_par.is_segment_width_default
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.segment_width == 3

    leg_par.segment_width = 2
    assert leg_par.segment_width == 2

    with pytest.raises(Exception):
        leg_par = LegendParameters(segment_width=0)
    with pytest.raises(Exception):
        leg_par.segment_width = 0


def test_text_height():
    """Test the LegendParameter text_height property."""
    leg_par = LegendParameters()
    leg_par.text_height = 0.25

    assert leg_par.text_height == 0.25
    assert not leg_par.is_text_height_default
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.text_height == 0.25

    leg_par.text_height = 2
    assert leg_par.text_height == 2

    with pytest.raises(Exception):
        leg_par = LegendParameters(text_height=0)
    with pytest.raises(Exception):
        leg_par.text_height = 0


def test_font():
    """Test the LegendParameter font property."""
    leg_par = LegendParameters()
    leg_par.font = 'Times'

    assert leg_par.font == 'Times'
    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.font == 'Times'

    leg_par.font = 'Courier'
    assert leg_par.font == 'Courier'

    with pytest.raises(Exception):
        leg_par = LegendParameters(font=0)
    with pytest.raises(Exception):
        leg_par.font = 0


def test_init_legend():
    """Test the initialization of Legend objects."""
    legend = Legend([0, 10])
    str(legend)  # Test the Legend representation

    assert len(legend) == 2
    assert legend[0] == 0
    assert legend[1] == 10
    for item in legend:
        assert isinstance(item, (float, int))

    assert len(legend.values) == 2
    assert legend.legend_parameters.min == 0
    assert legend.legend_parameters.max == 10
    assert legend.is_min_default
    assert legend.is_max_default
    assert legend.legend_parameters.is_segment_count_default
    assert legend.legend_parameters.is_title_default
    assert legend.legend_parameters.is_base_plane_default
    assert legend.legend_parameters.is_segment_height_default
    assert legend.legend_parameters.is_segment_width_default
    assert legend.legend_parameters.is_text_height_default

    legend_copy = legend.duplicate()
    assert legend_copy.values == legend.values
    assert legend_copy.legend_parameters.min == legend.legend_parameters.min
    assert legend_copy.legend_parameters.max == legend.legend_parameters.max
    assert legend_copy.legend_parameters.segment_count == \
        legend.legend_parameters.segment_count
    assert legend_copy.is_min_default
    assert legend_copy.is_max_default
    assert legend_copy.legend_parameters.is_segment_count_default
    assert legend_copy.legend_parameters.is_title_default
    assert legend_copy.legend_parameters.is_base_plane_default
    assert legend_copy.legend_parameters.is_segment_height_default
    assert legend_copy.legend_parameters.is_segment_width_default
    assert legend_copy.legend_parameters.is_text_height_default


def test_init_legend_with_parameter():
    """Test the initialization of Legend with LegendParameter objects."""
    legend = Legend([0, 10], LegendParameters(2, 8))

    assert len(legend.values) == 2
    assert legend.legend_parameters.min == 2
    assert legend.legend_parameters.max == 8
    assert not legend.is_min_default
    assert not legend.is_max_default


def test_legend_to_from_dict():
    """Test the to/from dict methods."""
    legend = Legend([0, 10], LegendParameters(2, 8))
    legend_dict = legend.to_dict()
    new_legend = Legend.from_dict(legend_dict)
    assert new_legend.to_dict() == legend_dict


def test_legend_value_colors():
    """Test the color_range, value_colors, and segment_colors property."""
    legend = Legend(range(5), LegendParameters(segment_count=10))

    for i, color in enumerate(legend.color_range):
        assert color == ColorRange()[i]
    assert legend.color_range.domain == ColorRange(domain=(0, 4)).domain

    assert len(legend.values) == 5
    assert len(legend.value_colors) == 5
    assert legend.value_colors[0] == Colorset.original()[0]
    assert legend.segment_colors == Colorset.original()


def test_legend_title():
    """Test the title property."""
    legend = Legend(range(5))
    assert legend.title == ''
    assert legend.legend_parameters.is_title_default
    assert legend.title_location.o == Point3D(0, 11.25, 0)
    assert legend.title_location_2d == Point2D(0, 11.25)

    legend = Legend(range(5), LegendParameters(segment_count=5, title='C'))
    assert legend.title == 'C'
    assert not legend.legend_parameters.is_title_default
    assert legend.title_location.o == Point3D(0, 5.25, 0)
    assert legend.title_location_2d == Point2D(0, 5.25)


def test_segment_text():
    """Test the segment_text property."""
    legend = Legend(range(10), LegendParameters(segment_count=6))

    assert len(legend.segment_text_location) == len(legend.segment_text) == 6
    assert legend.segment_text == ['0.00', '1.80', '3.60', '5.40', '7.20', '9.00']
    for pl in legend.segment_text_location:
        assert isinstance(pl, Plane)
    for pt in legend.segment_text_location_2d:
        assert isinstance(pt, Point2D)
    seg_num = [0, 1.8, 3.6, 5.4, 7.2, 9]
    for i, num in enumerate(legend.segment_numbers):
        assert num == pytest.approx(seg_num[i], rel=1e-3)

    legend = Legend(range(10), LegendParameters(segment_count=6))
    legend.legend_parameters.vertical = False

    assert len(legend.segment_text_location) == len(legend.segment_text) == 6
    assert legend.segment_text == ['0.00', '1.80', '3.60', '5.40', '7.20', '9.00']
    for pl in legend.segment_text_location:
        assert isinstance(pl, Plane)


def test_segment_text_ordinal_dictionary():
    """Test the segment_text property with ordinal dictionary."""
    results = list(range(3000))
    ordinal_dict = {300: 'low', 1150: 'desired', 2000: 'too much'}
    legend = Legend(results, LegendParameters(min=300, max=2000, segment_count=3))
    legend.legend_parameters.ordinal_dictionary = ordinal_dict

    assert len(legend.segment_text_location) == len(legend.segment_text) == 3
    assert legend.segment_text == ['low', 'desired', 'too much']
    assert len(set(legend.value_colors)) > 100


def test_segment_text_ordinal_dictionary_large():
    """Test the segment_text property with another larger ordinal dictionary."""
    results = list(range(-4, 5))
    ordinal_dict = PredictedMeanVote().unit_descr
    legend = Legend(results, LegendParameters(min=-3, max=3, segment_count=7))
    legend.legend_parameters.ordinal_dictionary = ordinal_dict
    assert legend.segment_text == ['Cold', 'Cool', 'Slightly Cool', 'Neutral',
                                   'Slightly Warm', 'Warm', 'Hot']

    legend.legend_parameters.min = -2
    legend.legend_parameters.max = 2
    legend.legend_parameters.segment_count = 5
    assert legend.segment_text == ['Cool', 'Slightly Cool', 'Neutral',
                                   'Slightly Warm', 'Warm']


def test_segment_mesh():
    """Test the segment_mesh property."""
    legend = Legend(range(10), LegendParameters(segment_count=6))

    assert isinstance(legend.segment_mesh, Mesh3D)
    assert len(legend.segment_mesh.faces) == 6
    assert len(legend.segment_mesh.vertices) == 14
    legend.legend_parameters.vertical = False
    assert len(legend.segment_mesh.faces) == 6
    assert len(legend.segment_mesh.vertices) == 14

    legend.legend_parameters.continuous_legend = True
    assert len(legend.segment_mesh.faces) == 5
    assert len(legend.segment_mesh.vertices) == 12
    legend.legend_parameters.vertical = True
    assert len(legend.segment_mesh.faces) == 5
    assert len(legend.segment_mesh.vertices) == 12


def test_segment_mesh_2d():
    """Test the segment_mesh_2d property."""
    legend = Legend(range(10), LegendParameters(segment_count=6))

    assert isinstance(legend.segment_mesh_2d, Mesh2D)
    assert len(legend.segment_mesh_2d.faces) == 6
    assert len(legend.segment_mesh_2d.vertices) == 14
    legend.legend_parameters.vertical = False
    assert len(legend.segment_mesh_2d.faces) == 6
    assert len(legend.segment_mesh_2d.vertices) == 14

    legend.legend_parameters.continuous_legend = True
    assert len(legend.segment_mesh_2d.faces) == 5
    assert len(legend.segment_mesh_2d.vertices) == 12
    legend.legend_parameters.vertical = True
    assert len(legend.segment_mesh_2d.faces) == 5
    assert len(legend.segment_mesh_2d.vertices) == 12


def test_init_legend_parameters_categorized():
    """Test the initialization of LegendParametersCategorized objects."""
    leg_colors = [Color(0, 0, 255), Color(0, 255, 0), Color(255, 0, 0)]
    leg_par = LegendParametersCategorized([300, 2000], leg_colors)
    leg_par.decimal_count = 0
    str(leg_par)  # Test the LegendParametersCategorized representation

    assert leg_par.domain == (300, 2000)
    assert leg_par.colors == tuple(leg_colors)
    assert leg_par.category_names == ('<300', '300 - 2000', '>2000')
    assert leg_par.min == 300
    assert leg_par.max == 2000
    assert leg_par.is_segment_count_default
    assert leg_par.is_title_default
    assert leg_par.is_base_plane_default
    assert leg_par.is_segment_height_default
    assert leg_par.is_segment_width_default
    assert leg_par.is_text_height_default

    leg_par_copy = leg_par.duplicate()
    assert leg_par_copy.domain == leg_par.domain
    assert leg_par_copy.colors == leg_par.colors
    assert leg_par_copy.category_names == leg_par.category_names


def test_categorized_to_from_dict():
    """Test the LegendParametersCategorized to/from dict methods."""
    leg_colors = [Color(0, 0, 255), Color(0, 255, 0), Color(255, 0, 0)]
    leg_par = LegendParametersCategorized([300, 2000], leg_colors)
    leg_par_dict = leg_par.to_dict()
    new_leg_par = LegendParametersCategorized.from_dict(leg_par_dict)
    assert new_leg_par.to_dict() == leg_par_dict


def test_categorized_colors():
    """Test the LegendParametersCategorized colors property."""
    data = [100, 300, 500, 1000, 2000, 3000]
    leg_colors = [Color(0, 0, 255), Color(0, 255, 0), Color(255, 0, 0)]
    legend_par = LegendParametersCategorized([300, 2000], leg_colors)

    legend = Legend(data, legend_par)
    assert legend.segment_colors == tuple(leg_colors)
    assert legend.value_colors == \
        (Color(0, 0, 255), Color(0, 255, 0), Color(0, 255, 0), Color(0, 255, 0),
         Color(0, 255, 0), Color(255, 0, 0))

    legend.legend_parameters.continuous_colors = True
    assert legend.value_colors == \
        (Color(0, 0, 255), Color(0, 0, 255), Color(0, 60, 195), Color(0, 210, 45),
         Color(255, 0, 0), Color(255, 0, 0))


def test_categorized_category_names():
    """Test the LegendParametersCategorized category_names property."""
    data = [100, 300, 500, 1000, 2000, 3000]
    leg_colors = [Color(0, 0, 255), Color(0, 255, 0), Color(255, 0, 0)]
    legend_par = LegendParametersCategorized([300, 2000], leg_colors)
    cat_names = ['low', 'desired', 'too much']
    legend_par.category_names = cat_names
    legend = Legend(data, legend_par)

    assert legend_par.category_names == tuple(cat_names)
    assert legend.segment_text == tuple(cat_names)

    with pytest.raises(AssertionError):
        legend_par.category_names = ['low', 'desired', 'too much', 'not a category']



# coding=utf-8
from ladybug.skymodel import estimate_illuminance_from_irradiance, \
    dirint, disc, _get_dirint_coeffs

import pytest
import math


def test_estimate_illuminance_from_irradiance():
    """Test the accuracy of the estimate_illuminance_from_irradiance function."""
    # Test a typical daytime condition
    gh_ill, dn_ill, dh_ill, z_lum = estimate_illuminance_from_irradiance(
        60, 200 + 800 * math.sin(math.radians(60)), 800, 200, 15)
    assert gh_ill == pytest.approx(99415.2739, rel=1e-2)
    assert dn_ill == pytest.approx(78879.2958, rel=1e-2)
    assert dh_ill == pytest.approx(27609.2088, rel=1e-2)
    assert z_lum == pytest.approx(9056.0109, rel=1e-2)

    # test a typical nighttime condition
    gh_ill, dn_ill, dh_ill, z_lum = estimate_illuminance_from_irradiance(
        0, 0, 0, 0, 15)
    assert gh_ill == 0
    assert dn_ill == 0
    assert dh_ill == 0
    assert z_lum == 0


def test_dirint():
    """Test the accuracy of the dirint model against pvlib results."""
    dirint_result = dirint(
        [1038.62, 254.53], [90 - 10.567, 90 - 72.469],
        [175, 175], [93193., 93193.])
    assert dirint_result[0] == pytest.approx(868.8332, rel=1e-2)
    assert dirint_result[1] == pytest.approx(699.65011, rel=1e-2)


def test_dirint_tdew():
    """Test the accuracy of the dirint model against pvlib results."""
    dirint_result = dirint(
        [1038.62, 254.53], [90 - 10.567, 90 - 72.469],
        [175, 175], [93193., 93193.], temp_dew=[10, 10])

    assert dirint_result[0] == pytest.approx(882.1372, rel=1e-2)
    assert dirint_result[1] == pytest.approx(672.5589, rel=1e-2)


def test_dirint_no_delta_kt():
    """Test the accuracy of the dirint model against pvlib results."""
    dirint_result = dirint(
        [1038.62, 254.53], [90 - 10.567, 90 - 72.469], [175, 175],
        [93193., 93193.], use_delta_kt_prime=False)
    assert dirint_result[0] == pytest.approx(861.857, rel=1e-1)
    assert dirint_result[1] == pytest.approx(670.341, rel=1e-3)


def test_dirint_coeffs():
    """Test the dirint coefficients against pvlib results."""
    coeffs = _get_dirint_coeffs()
    assert coeffs[0][0][0][0] == 0.385230
    assert coeffs[0][1][2][1] == 0.229970
    assert coeffs[3][2][6][3] == 1.032260


def test_disc():
    """Test the accuracy of the disc model against pvlib results."""
    disc_result = disc(1000, 80, 1)
    assert disc_result[0] == pytest.approx(611.40543, rel=1e-2)
    assert disc_result[1] == pytest.approx(0.716088, rel=1e-3)
    assert disc_result[2] == pytest.approx(1.014825, rel=1e-3)

    disc_result = disc(200, 20, 150)
    assert disc_result[0] == pytest.approx(203.0609, rel=1e-2)
    assert disc_result[1] == pytest.approx(0.43897, rel=1e-3)
    assert disc_result[2] == pytest.approx(2.89994, rel=1e-3)


def test_disc_overirradiance():
    """Test overirradiance in the disc model."""
    disc_result = disc(3000, 90, 200)

    assert disc_result[0] == pytest.approx(872.544, rel=1e-2)
    assert disc_result[1] == pytest.approx(1.000, rel=1e-3)
    assert disc_result[2] == pytest.approx(0.999493933, rel=1e-3)
